#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.


class Undefined(object):
    """Defines an undefined state of a property."""
    
    
    def __str__(self):
        """Returns string representation."""
        
        return "UNDEFINED"
    
    
    def __repr__(self):
        """Gets debug string representation."""
        
        return self.__str__()
    
    
    def __hash__(self):
        """Defines hash."""
        
        return hash(str(self))
    
    
    def __bool__(self):
        """Defines response for zero-check."""
        
        return False
    
    
    def __eq__(self, other):
        """Defines equality comparer."""
        
        return other is self or other == self.__str__()
    
    
    def __lt__(self, other):
        """Defines lower comparer."""
        
        return other is not self
    
    
    def __le__(self, other):
        """Defines lower or equal comparer."""
        
        return True
    
    
    def __gt__(self, other):
        """Defines greater comparer."""
        
        return False
    
    
    def __ge__(self, other):
        """Defines greater or equal comparer."""
        
        return other is self


# init as singleton
UNDEF = Undefined()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. properties import *
from . formatter import Formatter


class FuncFormatter(Formatter):
    """
    This formatter tool uses custom function to format given values.
    
    Properties:
        
        func: callable
            Specifies a function to be used for custom formatting. The function
            is expected to have just one argument for the input value and
            should return formatted string.
    """
    
    func = FuncProperty(UNDEF)
    
    
    def format(self, value, *args, **kwargs):
        """
        Formats given value using custom function.
        
        Args:
            value: any
                Value to be formatted.
        
        Returns:
            str
                Formatted label.
        """
        
        return self.func(value)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for ticker helpers."""
    
    
    def test_split_eng(self):
        """Tests whether engineering splits are calculated correctly."""
        
        splits = pero.SPLITS_ENG
        
        parts = pero.formatters.split_value(12345, splits)
        model = {k: 0 for k in splits}
        model.update({"k": 12, "": 345})
        self.assertEqual(parts, model)
    
    
    def test_split_time(self):
        """Tests whether time splits are calculated correctly."""
        
        splits = pero.SPLITS_TIME
        
        parts = pero.formatters.split_value(12345.6789, splits)
        model = {k: 0 for k in splits}
        model.update({"h": 3, "m": 25, "s": 45, "ms": 678, "us": 900})
        self.assertEqual(parts, model)
    
    
    def test_split_bytes(self):
        """Tests whether bytes splits are calculated correctly."""
        
        splits = pero.SPLITS_BYTES
        
        parts = pero.formatters.split_value(1234567.89, splits)
        model = {k: 0 for k in splits}
        model.update({"M": 1, "k": 181, "": 647})
        self.assertEqual(parts, model)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import wx
from ... enums import *


WX_LINE_CAP = {
    LINE_CAP_BUTT: wx.CAP_BUTT,
    LINE_CAP_SQUARE: wx.CAP_PROJECTING,
    LINE_CAP_ROUND: wx.CAP_ROUND}

WX_LINE_JOIN = {
    LINE_JOIN_BEVEL: wx.JOIN_BEVEL,
    LINE_JOIN_MITER: wx.JOIN_MITER,
    LINE_JOIN_ROUND: wx.JOIN_ROUND}

WX_LINE_STYLE = {
    LINE_STYLE_CUSTOM: wx.PENSTYLE_USER_DASH,
    LINE_STYLE_SOLID: wx.PENSTYLE_SOLID,
    LINE_STYLE_DOTTED: DASH_VALUES_DOTTED,
    LINE_STYLE_DASHED: DASH_VALUES_DASHED,
    LINE_STYLE_DASHDOTTED: DASH_VALUES_DASHDOTTED}

WX_FILL_STYLE = {
    FILL_STYLE_TRANS: wx.BRUSHSTYLE_TRANSPARENT,
    FILL_STYLE_SOLID: wx.BRUSHSTYLE_SOLID}

WX_FILL_RULE = {
    FILL_RULE_EVENODD: wx.ODDEVEN_RULE,
    FILL_RULE_WINDING: wx.WINDING_RULE}

WX_FONT_FAMILY = {
    FONT_FAMILY_SERIF: wx.FONTFAMILY_ROMAN,
    FONT_FAMILY_SANS: wx.FONTFAMILY_SWISS,
    FONT_FAMILY_MONO: wx.FONTFAMILY_TELETYPE}

WX_FONT_STYLE = {
    FONT_STYLE_NORMAL: wx.FONTSTYLE_NORMAL,
    FONT_STYLE_ITALIC: wx.FONTSTYLE_ITALIC}

WX_FONT_WEIGHT = {
    FONT_WEIGHT_NORMAL: wx.FONTWEIGHT_NORMAL,
    FONT_WEIGHT_LIGHT: wx.FONTWEIGHT_LIGHT,
    FONT_WEIGHT_BOLD: wx.FONTWEIGHT_BOLD,
    FONT_WEIGHT_BLACK: wx.FONTWEIGHT_BOLD,
    FONT_WEIGHT_HEAVY: wx.FONTWEIGHT_BOLD,
    FONT_WEIGHT_SEMIBOLD: wx.FONTWEIGHT_BOLD,
    FONT_WEIGHT_MEDIUM: wx.FONTWEIGHT_NORMAL,
    FONT_WEIGHT_ULTRALIGHT: wx.FONTWEIGHT_LIGHT,
    FONT_WEIGHT_THIN: wx.FONTWEIGHT_LIGHT}

WX_CURSORS = {
    CURSOR_BLANK: wx.CURSOR_BLANK,
    CURSOR_ARROW: wx.CURSOR_ARROW,
    CURSOR_HAND: wx.CURSOR_HAND,
    CURSOR_SIZEWE: wx.CURSOR_SIZEWE,
    CURSOR_SIZENS: wx.CURSOR_SIZENS}

WX_KEYS = {
    wx.WXK_TAB: KEY_TAB,
    wx.WXK_RETURN: KEY_RETURN,
    
    wx.WXK_SHIFT: KEY_SHIFT,
    wx.WXK_CONTROL: KEY_CTRL,
    wx.WXK_ALT: KEY_ALT,
    
    wx.WXK_ESCAPE: KEY_ESC,
    wx.WXK_SPACE: KEY_SPACE,
    
    wx.WXK_PAGEUP: KEY_PAGEUP,
    wx.WXK_PAGEDOWN: KEY_PAGEDOWN,
    wx.WXK_END: KEY_END,
    wx.WXK_HOME: KEY_HOME,
    wx.WXK_LEFT: KEY_LEFT,
    wx.WXK_UP: KEY_UP,
    wx.WXK_RIGHT: KEY_RIGHT,
    wx.WXK_DOWN: KEY_DOWN,
    
    wx.WXK_DELETE: KEY_DELETE,
    
    wx.WXK_F1: KEY_F1,
    wx.WXK_F2: KEY_F2,
    wx.WXK_F3: KEY_F3,
    wx.WXK_F4: KEY_F4,
    wx.WXK_F5: KEY_F5,
    wx.WXK_F6: KEY_F6,
    wx.WXK_F7: KEY_F7,
    wx.WXK_F8: KEY_F8,
    wx.WXK_F9: KEY_F9,
    wx.WXK_F10: KEY_F10,
    wx.WXK_F11: KEY_F11,
    wx.WXK_F12: KEY_F12}

WX_RASTER_TYPES = {
    '.bmp': wx.BITMAP_TYPE_BMP,
    '.jpg': wx.BITMAP_TYPE_JPEG,
    '.jpeg': wx.BITMAP_TYPE_JPEG,
    '.png': wx.BITMAP_TYPE_PNG,
    '.pcx': wx.BITMAP_TYPE_PCX,
    '.pnm': wx.BITMAP_TYPE_PNM,
    '.tif': wx.BITMAP_TYPE_TIFF,
    '.tiff': wx.BITMAP_TYPE_TIFF,
    '.xpm': wx.BITMAP_TYPE_XPM,
    '.ico': wx.BITMAP_TYPE_ICO,
    '.cur': wx.BITMAP_TYPE_CUR}



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for bytes formatter."""
    
    
    def test_standalone(self):
        """Tests whether formatter works as standalone tool."""
        
        formatter = pero.BytesFormatter()
        
        self.assertEqual(formatter.format(234.5678e6), "224 MB")
        self.assertEqual(formatter.format(234.5678e3), "229 kB")
        self.assertEqual(formatter.format(234.5678), "235 B")
        self.assertEqual(formatter.format(234.5678e-3), "0 B")
    
    
    def test_places(self):
        """Tests whether places works correctly."""
        
        formatter = pero.BytesFormatter(places=0)
        
        self.assertEqual(formatter.format(234.5678e6), "224 MB")
        self.assertEqual(formatter.format(234.5678e3), "229 kB")
        self.assertEqual(formatter.format(234.5678), "235 B")
        self.assertEqual(formatter.format(234.5678e-3), "0 B")
        self.assertEqual(formatter.format(234.5678e-6), "0 B")
        
        formatter = pero.BytesFormatter(places=2)
        
        self.assertEqual(formatter.format(234.5678e6), "223.70 MB")
        self.assertEqual(formatter.format(234.5678e3), "229.07 kB")
        self.assertEqual(formatter.format(234.5678), "234.57 B")
        self.assertEqual(formatter.format(234.5678e-3), "0.23 B")
        self.assertEqual(formatter.format(234.5678e-6), "0.00 B")
    
    
    def test_precision(self):
        """Tests whether precision works correctly."""
        
        formatter = pero.BytesFormatter(precision=200)
        
        self.assertEqual(formatter.format(234.5678e6), "223.7013 MB")
        self.assertEqual(formatter.format(234.5678e3), "229.1 kB")
        self.assertEqual(formatter.format(234.5678), "235 B")
        self.assertEqual(formatter.format(234.5678e-3), "0 B")
        self.assertEqual(formatter.format(234.5678e-6), "0 B")
        
        formatter = pero.BytesFormatter(precision=0.2)
        
        self.assertEqual(formatter.format(234.5678e6), "223.7012863 MB")
        self.assertEqual(formatter.format(234.5678e3), "229.0701 kB")
        self.assertEqual(formatter.format(234.5678), "234.6 B")
        self.assertEqual(formatter.format(234.5678e-3), "0.2 B")
        self.assertEqual(formatter.format(234.5678e-6), "0.0 B")
    
    
    def test_domain(self):
        """Tests whether domain works correctly."""
        
        formatter = pero.BytesFormatter(domain=1e4)
        
        self.assertEqual(formatter.format(234.5678e6), "229070 kB")
        self.assertEqual(formatter.format(234.5678e3), "229 kB")
        self.assertEqual(formatter.format(234.5678), "0 kB")
        self.assertEqual(formatter.format(234.5678e-3), "0 kB")
        self.assertEqual(formatter.format(234.5678e-6), "0 kB")
        
        formatter = pero.BytesFormatter(domain=1e2)
        
        self.assertEqual(formatter.format(234.5678e6), "234567800 B")
        self.assertEqual(formatter.format(234.5678e3), "234568 B")
        self.assertEqual(formatter.format(234.5678), "235 B")
        self.assertEqual(formatter.format(234.5678e-3), "0 B")
        self.assertEqual(formatter.format(234.5678e-6), "0 B")
    
    
    def test_digits(self):
        """Tests whether digits works correctly."""
        
        # check domain and precision
        formatter = pero.BytesFormatter(domain=1e4, precision=200)
        
        self.assertEqual(formatter.format(234.5678e6), "229070.1 kB")
        self.assertEqual(formatter.format(234.5678e3), "229.1 kB")
        self.assertEqual(formatter.format(234.5678), "0.2 kB")
        self.assertEqual(formatter.format(234.5678e-3), "0.0 kB")
        self.assertEqual(formatter.format(234.5678e-6), "0.0 kB")
        
        # check domain, precision and places
        formatter = pero.BytesFormatter(domain=1e4, precision=200, places=2)
        
        self.assertEqual(formatter.format(234.5678e6), "229070.12 kB")
        self.assertEqual(formatter.format(234.5678e3), "229.07 kB")
        self.assertEqual(formatter.format(234.5678), "0.23 kB")
        self.assertEqual(formatter.format(234.5678e-3), "0.00 kB")
        self.assertEqual(formatter.format(234.5678e-6), "0.00 kB")
    
    
    def test_suffix(self):
        """Tests whether suffix works correctly."""
        
        # standalone formatter
        formatter = pero.BytesFormatter(hide_suffix=True)
        self.assertEqual(formatter.suffix(), "")
        
        self.assertEqual(formatter.format(234.5678e6), "224")
        self.assertEqual(formatter.suffix(), " (MB)")
        
        self.assertEqual(formatter.format(234.5678e3), "229")
        self.assertEqual(formatter.suffix(), " (kB)")
        
        self.assertEqual(formatter.format(234.5678), "235")
        self.assertEqual(formatter.suffix(), " (B)")
        
        # formatter with domain
        formatter = pero.BytesFormatter(hide_suffix=True, domain=1e4)
        self.assertEqual(formatter.suffix(), " (kB)")
        
        self.assertEqual(formatter.format(234.5678e6), "229070")
        self.assertEqual(formatter.suffix(), " (kB)")
        
        self.assertEqual(formatter.format(234.5678e3), "229")
        self.assertEqual(formatter.suffix(), " (kB)")
        
        self.assertEqual(formatter.format(234.5678), "0")
        self.assertEqual(formatter.suffix(), " (kB)")
    
    
    def test_convert(self):
        """Tests whether convert works correctly."""
        
        formatter = pero.BytesFormatter(domain=1e4)
        
        self.assertEqual(formatter.convert(234.5678e6), 229070.1171875)
        self.assertEqual(formatter.convert(234.5678e4), 2290.701171875)
        self.assertEqual(formatter.convert(234.5678), 0.2290701171875)
        self.assertEqual(formatter.convert(234.5678e-4), 2.290701171875e-05)
        self.assertEqual(formatter.convert(234.5678e-6), 2.290701171875e-07)
    
    
    def test_invert(self):
        """Tests whether invert works correctly."""
        
        formatter = pero.BytesFormatter(domain=1e4)
        
        self.assertEqual(formatter.invert(229070.1171875), 234.5678e6)
        self.assertEqual(formatter.invert(2290.701171875), 234.5678e4)
        self.assertEqual(formatter.invert(0.2290701171875), 234.5678)
        self.assertEqual(formatter.invert(2.290701171875e-05), 234.5678e-4)
        self.assertEqual(formatter.invert(2.290701171875e-07), 234.5678e-6)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import os.path
import wx
from . enums import *
from . canvas import WXCanvas
from . viewer import WXViewer


def show(graphics, title=None, width=None, height=None, **options):
    """
    Shows given graphics in the viewer app.
    
    Args:
        graphics: pero.Graphics or pero.Control
            Graphics to be shown.
        
        title: str or None
            Viewer frame title.
        
        width: float or None
            Viewer width in device units.
        
        height: float or None
            Viewer height in device units.
    """
    
    # init app
    app = WXViewer(0)
    
    # set title
    if title is not None:
        app.set_title(title)
    
    # check size
    if not width:
        width = VIEWER_WIDTH
    if not height:
        height = VIEWER_HEIGHT
    
    # set size
    app.set_size((width, height))
    
    # set graphics
    app.set_content(graphics)
    
    # draw graphics
    app.refresh()
    
    # start app
    app.show()


def export(graphics, path, width=None, height=None, **options):
    """
    Draws given graphics as raster image into specified file. The image format
    is determined from the extension of given file path. Supported extensions
    are .bmp, .jpg, .jpeg, .png, .pcx, .pnm, .tif, .tiff, .xpm, .ico and .cur.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str
            Full path of a file to save the image into.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        draw_scale: float
            Drawing scaling factor.
        
        line_scale: float
            Line scaling factor.
        
        font_scale: float
            Font scaling factor.
        
        dpi: int
            Image resolution as dots-per-inch.
        
        quality: int
            Image quality in range between 0 and 100 with 0 meaning very poor
            and 100 excellent. This option is only available for JPEG format.
    """
    
    # get filename and extension
    dirname, filename = os.path.split(path)
    basename, extension = os.path.splitext(filename)
    extension = extension.lower()
    
    # check format
    if extension not in WX_RASTER_TYPES:
        message = "Unsupported image format! -> %s" % extension
        raise NotImplementedError(message)
    
    # init simple app if necessary
    try:
        app = wx.App()
    except:
        pass
    
    # check size
    if not width:
        width = EXPORT_WIDTH
    if not height:
        height = EXPORT_HEIGHT
    
    # create DC
    bitmap = wx.Bitmap(width, height)
    dc = wx.MemoryDC()
    dc.SelectObject(bitmap)
    
    # use GCDC
    if 'wxMac' not in wx.PlatformInfo:
        dc = wx.GCDC(dc)
    
    # init canvas
    canvas = WXCanvas(dc, width=width, height=height)
    
    if 'draw_scale' in options:
        canvas.draw_scale = options['draw_scale']
    
    if 'line_scale' in options:
        canvas.line_scale = options['line_scale']
    
    if 'font_scale' in options:
        canvas.font_scale = options['font_scale']
    
    # draw graphics
    graphics.draw(canvas)
    
    # get image
    image = bitmap.ConvertToImage()
    
    # set image options
    if 'dpi' in options:
        image.SetOption(wx.IMAGE_OPTION_RESOLUTIONUNIT, str(wx.IMAGE_RESOLUTION_INCHES))
        image.SetOption(wx.IMAGE_OPTION_RESOLUTION, str(options['dpi']))
    
    if 'quality' in options:
        image.SetOption(wx.IMAGE_OPTION_QUALITY, str(options['quality']))
    else:
        image.SetOption(wx.IMAGE_OPTION_QUALITY, '95')
    
    # save to file
    image.SaveFile(path, WX_RASTER_TYPES[extension])



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# import main objects
from . interpols import Interpol, LinInterpol, LogInterpol, PowInterpol

from . scale import Scale
from . ordinal import OrdinalScale
from . levels import LevelScale
from . continuous import ContinuousScale, LinScale, LogScale, PowScale
from . gradient import GradientScale, GradientLinScale, GradientLogScale, GradientPowScale
from . converter import Converter



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from . glyph import Glyph
from . shapes import Wedge


class Gauge(Glyph):
    """
    Abstract base class for various types of gauges. It provides a simple glyph
    to visualize specific portion within a context of full data range.
    
    The 'start' and 'end' positions of the selected range are given as relative
    values within the range as %/100.
    
    By default the range increases the same way as device units, i.e. from left
    to right, from top to bottom or clockwise. This behavior can be changed by
    setting the 'reverse' property to True.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the origin.
        
        y: int, float or callable
            Specifies the y-coordinate of the origin.
        
        start: float or callable
            Specifies the start position of the foreground as a relative value
            of the full length in %/100."
        
        end: float or callable
            Specifies the end position of the foreground as a relative value
            of the full length in %/100."
        
        limit: int, float or callable
            Specifies the minimum display length of the foreground.
        
        reverse: bool or callable
            Specifies whether the foreground is drawn considering the top/left
            side (False) or the bottom/right side (True) as its start.
        
        bgr_line properties:
            Includes pero.LineProperties to specify the background outline.
        
        bgr_fill properties:
            Includes pero.FillProperties to specify the background fill.
        
        for_line properties:
            Includes pero.LineProperties to specify the foreground outline.
        
        for_fill properties:
            Includes pero.FillProperties to specify the foreground fill.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    
    start = NumProperty(0)
    end = NumProperty(0)
    limit = NumProperty(0)
    reverse = BoolProperty(False)
    
    bgr_pen = Include(LineProperties, prefix="bgr_", line_color="#000")
    bgr_fill = Include(FillProperties, prefix="bgr_", fill_color="#ddd")
    
    for_pen = Include(LineProperties, prefix="for_", line_color="#000")
    for_fill = Include(FillProperties, prefix="for_", fill_color="#000")


class StraitGauge(Gauge):
    """
    Strait gauge is drawn as a horizontal or vertical bar of specified length
    and thickness.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the top-left corner.
        
        y: int, float or callable
            Specifies the y-coordinate of the top-left corner.
        
        length: int, float or callable
            Specifies the bar length.
        
        thickness: int, float or callable
            Specifies the bar thickness.
        
        radius: int, float, tuple, callable or UNDEF
            Specifies the corner radius as a single value or values for
            individual corners starting from top-left.
        
        orientation: pero.ORIENTATION or callable
            Specifies the bar orientation as any item from the
            pero.ORIENTATION enum.
    """
    
    length = NumProperty(0)
    thickness = NumProperty(7)
    radius = QuadProperty(0)
    orientation = EnumProperty(ORI_HORIZONTAL, enum=ORIENTATION)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw the gauge."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        length = self.get_property('length', source, overrides)
        thickness = self.get_property('thickness', source, overrides)
        radius = self.get_property('radius', source, overrides)
        limit = self.get_property('limit', source, overrides)
        orientation = self.get_property('orientation', source, overrides)
        reverse = self.get_property('reverse', source, overrides)
        start = self.get_property('start', source, overrides)
        end = self.get_property('end', source, overrides)
        
        # check edges
        start = max(0, min(1, start))
        end = max(0, min(1, end))
        
        # reverse direction
        if reverse:
            start = 1 - start
            end = 1 - end
        
        # check direction
        if start > end:
            start, end = end, start
        
        # make highlight absolute
        start = length*start
        end = length*end
        
        # check highlight limit
        if radius:
            limit = max(limit, 2*max(radius))
        
        if abs(end - start) < limit:
            diff = limit - abs(end - start)
            
            if start > .5*diff and end < length-.5*diff:
                start -= .5*diff
                end += .5*diff
            
            elif start > diff:
                start -= diff
            
            elif end < length-diff:
                end += diff
        
        # finalize by orientation
        if orientation == ORI_HORIZONTAL:
            bgr_width = length
            bgr_height = thickness
            for_x = x + start
            for_y = y
            for_width = end - start
            for_height = thickness
        
        else:
            bgr_width = thickness
            bgr_height = length
            for_x = x
            for_y = y + start
            for_width = thickness
            for_height = end - start
        
        # start drawing group
        canvas.group(tag, "gauge")
        
        # set pen and brush for background
        canvas.set_pen_by(self, prefix="bgr_", source=source, overrides=overrides)
        canvas.set_brush_by(self, prefix="bgr_", source=source, overrides=overrides)
        
        # draw background
        canvas.draw_rect(x, y, bgr_width, bgr_height, radius)
        
        # set pen and brush for range
        canvas.set_pen_by(self, prefix="for_", source=source, overrides=overrides)
        canvas.set_brush_by(self, prefix="for_", source=source, overrides=overrides)
        
        # draw range
        canvas.draw_rect(for_x, for_y, for_width, for_height, radius)
        
        # end drawing group
        canvas.ungroup()


class RadialGauge(Gauge):
    """
    Radial gauge is drawn as a circle or an arc.
    
    Properties:
        
        inner_radius: int, float or callable
            Specifies the inner radius.
        
        outer_radius: int, float or callable
            Specifies the outer radius.
        
        start_angle properties:
            Includes pero.AngleProperties to specify the start angle.
        
        end_angle properties:
            Includes pero.AngleProperties to specify the end angle.
        
        clockwise: bool or callable
            Specifies the drawing direction. If set to True the arc is drawn
            clockwise, otherwise anti-clockwise.
    """
    
    inner_radius = NumProperty(UNDEF)
    outer_radius = NumProperty(UNDEF)
    start_angle = Include(AngleProperties, prefix="start_")
    end_angle = Include(AngleProperties, prefix="end_")
    clockwise = BoolProperty(True)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw the gauge."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        bgr_start = AngleProperties.get_angle(self, 'start_', ANGLE_RAD, source, overrides)
        bgr_end = AngleProperties.get_angle(self, 'end_', ANGLE_RAD, source, overrides)
        inner_radius = self.get_property('inner_radius', source, overrides)
        outer_radius = self.get_property('outer_radius', source, overrides)
        clockwise = self.get_property('clockwise', source, overrides)
        limit = self.get_property('limit', source, overrides)
        reverse = self.get_property('reverse', source, overrides)
        start = self.get_property('start', source, overrides)
        end = self.get_property('end', source, overrides)
        
        # check edges
        start = max(0, min(1, start))
        end = max(0, min(1, end))
        
        # reverse direction
        if reverse:
            start = 1 - start
            end = 1 - end
        
        # check direction
        if start > end:
            start, end = end, start
        
        # get full range
        length = (bgr_end - bgr_start) % (2*math.pi)
        if not clockwise:
            length -= 2*math.pi
        
        if length == 0 and bgr_start != bgr_end:
            length = 2*math.pi
        
        # get highlight angles
        for_start = length*start
        for_end = length*end
        
        # finalize angles
        for_start = bgr_start + for_start
        for_end = bgr_start + for_end
        
        # init glyph
        glyph = Wedge(
            x = x,
            y = y,
            inner_radius = inner_radius,
            outer_radius = outer_radius,
            start_angle_units = ANGLE_RAD,
            end_angle_units = ANGLE_RAD,
            clockwise = clockwise)
        
        # start drawing group
        canvas.group(tag, "gauge")
        
        # draw background
        glyph.set_properties_from(self, src_prefix="bgr_", source=source, overrides=overrides)
        glyph.draw(canvas, start_angle=bgr_start, end_angle=bgr_end)
        
        # draw range
        glyph.set_properties_from(self, src_prefix="for_", source=source, overrides=overrides)
        glyph.draw(canvas, start_angle=for_start, end_angle=for_end)
        
        # end drawing group
        canvas.ungroup()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for scalar formatter."""
    
    
    def test_auto_range(self):
        """Tests whether domain together with precision are used correctly."""
        
        formatter = pero.ScalarFormatter(
            domain = 7e6,
            precision = 100,
            sci_notation = False)
        
        self.assertEqual(formatter.format(0), "0")
        self.assertEqual(formatter.format(0.123456), "0")
        self.assertEqual(formatter.format(1.23456), "1")
        self.assertEqual(formatter.format(123.456), "123")
        self.assertEqual(formatter.format(1234.56), "1235")
        
        formatter = pero.ScalarFormatter(
            domain = 7e3,
            precision = 0.5,
            sci_notation = False)
        
        self.assertEqual(formatter.format(0), "0.0")
        self.assertEqual(formatter.format(0.123456), "0.1")
        self.assertEqual(formatter.format(1.23456), "1.2")
        self.assertEqual(formatter.format(123.456), "123.5")
        self.assertEqual(formatter.format(1234.56), "1234.6")
        
        formatter = pero.ScalarFormatter(
            domain = 7e-3,
            precision = 0.0005,
            sci_notation = False)
        
        self.assertEqual(formatter.format(0), "0.0000")
        self.assertEqual(formatter.format(0.123456), "0.1235")
        self.assertEqual(formatter.format(1.23456), "1.2346")
        self.assertEqual(formatter.format(123.456), "123.4560")
        self.assertEqual(formatter.format(1234.56), "1234.5600")
    
    
    def test_sci_notation(self):
        """Tests whether scientific notation works correctly."""
        
        formatter = pero.ScalarFormatter(
            domain = 7e6,
            precision = 100,
            sci_notation = True,
            sci_threshold = 3)
        
        self.assertEqual(formatter.format(0), "0e+00")
        self.assertEqual(formatter.format(0.123456), "1e-01")
        self.assertEqual(formatter.format(1.23456), "1e+00")
        self.assertEqual(formatter.format(123.456), "1e+02")
        self.assertEqual(formatter.format(1234.56), "1.2e+03")
        
        formatter = pero.ScalarFormatter(
            domain = 7e3,
            precision = 0.1,
            sci_notation = True,
            sci_threshold = 3)
        
        self.assertEqual(formatter.format(0), "0e+00")
        self.assertEqual(formatter.format(0.123456), "1e-01")
        self.assertEqual(formatter.format(1.23456), "1.2e+00")
        self.assertEqual(formatter.format(123.456), "1.235e+02")
        self.assertEqual(formatter.format(1234.56), "1.2346e+03")
        
        formatter = pero.ScalarFormatter(
            domain = 7e-3,
            precision = 0.001,
            sci_notation = True,
            sci_threshold = 3)
        
        self.assertEqual(formatter.format(0), "0e+00")
        self.assertEqual(formatter.format(0.123456), "1.23e-01")
        self.assertEqual(formatter.format(1.23456), "1.235e+00")
        self.assertEqual(formatter.format(123.456), "1.23456e+02")
        self.assertEqual(formatter.format(1234.56), "1.234560e+03")
    
    
    def test_suffix(self):
        """Tests whether suffix works correctly."""
        
        formatter = pero.ScalarFormatter(
            domain = 7e6,
            precision = 100,
            sci_notation = True,
            sci_threshold = 3,
            hide_suffix = True,
            suffix_template = "e{:.0f}")
        
        self.assertEqual(formatter.suffix(), "e6")
        
        self.assertEqual(formatter.format(0), "0.0000")
        self.assertEqual(formatter.format(0.123456), "0.0000")
        self.assertEqual(formatter.format(1.23456), "0.0000")
        self.assertEqual(formatter.format(123.456), "0.0001")
        self.assertEqual(formatter.format(1234.56), "0.0012")
        
        formatter = pero.ScalarFormatter(
            domain = 7e3,
            precision = 0.1,
            sci_notation = True,
            sci_threshold = 3,
            hide_suffix = True,
            suffix_template = "e{:.0f}")
        
        self.assertEqual(formatter.suffix(), "e3")
        
        self.assertEqual(formatter.format(0), "0.0000")
        self.assertEqual(formatter.format(0.123456), "0.0001")
        self.assertEqual(formatter.format(1.23456), "0.0012")
        self.assertEqual(formatter.format(123.456), "0.1235")
        self.assertEqual(formatter.format(1234.56), "1.2346")
        
        formatter = pero.ScalarFormatter(
            domain = 7e-3,
            precision = 0.001,
            sci_notation = True,
            sci_threshold = 3,
            hide_suffix = True,
            suffix_template = "e{:.0f}")
        
        self.assertEqual(formatter.suffix(), "e-3")
        
        self.assertEqual(formatter.format(0), "0")
        self.assertEqual(formatter.format(0.123456), "123")
        self.assertEqual(formatter.format(1.23456), "1235")
        self.assertEqual(formatter.format(123.456), "123456")
        self.assertEqual(formatter.format(1234.56), "1234560")


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import math


def calc_step_size(domain, count=7, splits=(5, 3, 2, 1), base=10):
    """
    Calculates visually nice step size for given range.
    
    Args:
        domain: int or float
            Absolute value of current range.
        
        count: int
            Expected number of ticks.
        
        splits: (float,)
            Allowed splits.
        
        base: int
            Log base.
    
    Returns:
        float
            Step size.
    """
    
    ideal = math.fabs(domain) / float(max(1, count))
    log = math.log(ideal, base)
    power = math.floor(log)
    fraction = log - power
    error = float('inf')  # fraction
    
    factor = 1.
    for split in splits:
        e = math.fabs(fraction - math.log(split, base))
        if e < error:
            factor = split
            error = e
    
    step = factor * math.pow(base, power)
    
    return step


def make_lin_ticks(start, end, step):
    """
    Makes equidistant ticks for given range and step size.
    
    Args:
        start: int or float
            Minimum value of the range.
        
        end: int or float
            Maximum value of the range.
        
        step: float
            Step size.
    
    Returns:
        (float,)
            Ticks values.
    """
    
    # init buffer
    ticks = []
    
    # ascending
    if start < end:
        tick = math.ceil(float(start) / step) * step
        while tick <= end:
            ticks.append(tick)
            tick += step
    
    # descending
    else:
        tick = math.floor(float(start) / step) * step
        while tick >= end:
            ticks.append(tick)
            tick -= step
    
    return tuple(ticks)


def make_log_ticks(start, end, base):
    """
    Makes logarithmic ticks for given range and base.
    
    Args:
        start: int or float
            Minimum value of the range.
        
        end: int or float
            Maximum value of the range.
        
        base: float
            Logarithm base.
    
    Returns:
        (float,)
            Ticks values.
    """
    
    ticks = []
    
    lo = round(math.log(start, base)) - 1
    hi = round(math.log(end, base)) + 1
    
    while lo < hi:
        
        p = math.pow(base, lo)
        lo += 1
        
        b = 1
        while b < base:
            
            t = p*b
            b += 1
            
            if t < start:
                continue
            if t > end:
                break
            
            ticks.append(t)
    
    return ticks



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# import main objects
from . ticker import Ticker
from . linear import LinTicker
from . logarithmic import LogTicker
from . fixed import FixTicker
from . timed import TimeTicker

# import utils
from . utils import *



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . enums import *
from . sketch import ShowSketch, ExportSketch


def show(graphics, title=None, width=None, height=None, **options):
    """
    Shows given graphics in the viewer app.
    
    Args:
        graphics: pero.Graphics or pero.Control
            Graphics to be shown.
        
        title: str or None
            Viewer frame title.
        
        width: float or None
            Viewer width in device units.
        
        height: float or None
            Viewer height in device units.
    """
    
    # check size
    if not width:
        width = EXPORT_WIDTH
    if not height:
        height = EXPORT_HEIGHT
    
    # init sketch
    sketch = ShowSketch(graphics, title, width, height, **options)
    
    # show sketch
    sketch.run_sketch()


def export(graphics, path, width=None, height=None, **options):
    """
    Draws given graphics as PDF document into specified file.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str
            Full path of a file to save the image into.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        draw_scale: float
            Drawing scaling factor.
        
        line_scale: float
            Line scaling factor.
        
        font_scale: float
            Font scaling factor.
    """
    
    # check size
    if not width:
        width = EXPORT_WIDTH
    if not height:
        height = EXPORT_HEIGHT
    
    # init sketch
    sketch = ExportSketch(graphics, path, width, height, **options)
    
    # draw and save to file
    sketch.run_sketch()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . color import Color
from . palette import Palette
from . gradient import Gradient


# define colors
Transparent = Color('#FFF0', name='Transparent')
AliceBlue = Color('#F0F8FF', name='AliceBlue')
AntiqueWhite = Color('#FAEBD7', name='AntiqueWhite')
Aqua = Color('#00FFFF', name='Aqua')
Aquamarine = Color('#7FFFD4', name='Aquamarine')
Azure = Color('#F0FFFF', name='Azure')
Beige = Color('#F5F5DC', name='Beige')
Bisque = Color('#FFE4C4', name='Bisque')
Black = Color('#000000', name='Black')
BlanchedAlmond = Color('#FFEBCD', name='BlanchedAlmond')
Blue = Color('#0000FF', name='Blue')
BlueViolet = Color('#8A2BE2', name='BlueViolet')
Brown = Color('#A52A2A', name='Brown')
BurlyWood = Color('#DEB887', name='BurlyWood')
CadetBlue = Color('#5F9EA0', name='CadetBlue')
Chartreuse = Color('#7FFF00', name='Chartreuse')
Chocolate = Color('#D2691E', name='Chocolate')
Coral = Color('#FF7F50', name='Coral')
CornflowerBlue = Color('#6495ED', name='CornflowerBlue')
Cornsilk = Color('#FFF8DC', name='Cornsilk')
Crimson = Color('#DC143C', name='Crimson')
Cyan = Color('#00FFFF', name='Cyan')
DarkBlue = Color('#00008B', name='DarkBlue')
DarkCyan = Color('#008B8B', name='DarkCyan')
DarkGoldenRod = Color('#B8860B', name='DarkGoldenRod')
DarkGray = Color('#A9A9A9', name='DarkGray')
DarkGrey = Color('#A9A9A9', name='DarkGrey')
DarkGreen = Color('#006400', name='DarkGreen')
DarkKhaki = Color('#BDB76B', name='DarkKhaki')
DarkMagenta = Color('#8B008B', name='DarkMagenta')
DarkOliveGreen = Color('#556B2F', name='DarkOliveGreen')
DarkOrange = Color('#FF8C00', name='DarkOrange')
DarkOrchid = Color('#9932CC', name='DarkOrchid')
DarkRed = Color('#8B0000', name='DarkRed')
DarkSalmon = Color('#E9967A', name='DarkSalmon')
DarkSeaGreen = Color('#8FBC8F', name='DarkSeaGreen')
DarkSlateBlue = Color('#483D8B', name='DarkSlateBlue')
DarkSlateGray = Color('#2F4F4F', name='DarkSlateGray')
DarkSlateGrey = Color('#2F4F4F', name='DarkSlateGrey')
DarkTurquoise = Color('#00CED1', name='DarkTurquoise')
DarkViolet = Color('#9400D3', name='DarkViolet')
DeepPink = Color('#FF1493', name='DeepPink')
DeepSkyBlue = Color('#00BFFF', name='DeepSkyBlue')
DimGray = Color('#696969', name='DimGray')
DimGrey = Color('#696969', name='DimGrey')
DodgerBlue = Color('#1E90FF', name='DodgerBlue')
FireBrick = Color('#B22222', name='FireBrick')
FloralWhite = Color('#FFFAF0', name='FloralWhite')
ForestGreen = Color('#228B22', name='ForestGreen')
Fuchsia = Color('#FF00FF', name='Fuchsia')
Gainsboro = Color('#DCDCDC', name='Gainsboro')
GhostWhite = Color('#F8F8FF', name='GhostWhite')
Gold = Color('#FFD700', name='Gold')
GoldenRod = Color('#DAA520', name='GoldenRod')
Gray = Color('#808080', name='Gray')
Grey = Color('#808080', name='Grey')
Green = Color('#008000', name='Green')
GreenYellow = Color('#ADFF2F', name='GreenYellow')
HoneyDew = Color('#F0FFF0', name='HoneyDew')
HotPink = Color('#FF69B4', name='HotPink')
IndianRed = Color('#CD5C5C', name='IndianRed')
Indigo = Color('#4B0082', name='Indigo')
Ivory = Color('#FFFFF0', name='Ivory')
Khaki = Color('#F0E68C', name='Khaki')
Lavender = Color('#E6E6FA', name='Lavender')
LavenderBlush = Color('#FFF0F5', name='LavenderBlush')
LawnGreen = Color('#7CFC00', name='LawnGreen')
LemonChiffon = Color('#FFFACD', name='LemonChiffon')
LightBlue = Color('#ADD8E6', name='LightBlue')
LightCoral = Color('#F08080', name='LightCoral')
LightCyan = Color('#E0FFFF', name='LightCyan')
LightGoldenRodYellow = Color('#FAFAD2', name='LightGoldenRodYellow')
LightGray = Color('#D3D3D3', name='LightGray')
LightGrey = Color('#D3D3D3', name='LightGrey')
LightGreen = Color('#90EE90', name='LightGreen')
LightPink = Color('#FFB6C1', name='LightPink')
LightSalmon = Color('#FFA07A', name='LightSalmon')
LightSeaGreen = Color('#20B2AA', name='LightSeaGreen')
LightSkyBlue = Color('#87CEFA', name='LightSkyBlue')
LightSlateGray = Color('#778899', name='LightSlateGray')
LightSlateGrey = Color('#778899', name='LightSlateGrey')
LightSteelBlue = Color('#B0C4DE', name='LightSteelBlue')
LightYellow = Color('#FFFFE0', name='LightYellow')
Lime = Color('#00FF00', name='Lime')
LimeGreen = Color('#32CD32', name='LimeGreen')
Linen = Color('#FAF0E6', name='Linen')
Magenta = Color('#FF00FF', name='Magenta')
Maroon = Color('#800000', name='Maroon')
MediumAquaMarine = Color('#66CDAA', name='MediumAquaMarine')
MediumBlue = Color('#0000CD', name='MediumBlue')
MediumOrchid = Color('#BA55D3', name='MediumOrchid')
MediumPurple = Color('#9370DB', name='MediumPurple')
MediumSeaGreen = Color('#3CB371', name='MediumSeaGreen')
MediumSlateBlue = Color('#7B68EE', name='MediumSlateBlue')
MediumSpringGreen = Color('#00FA9A', name='MediumSpringGreen')
MediumTurquoise = Color('#48D1CC', name='MediumTurquoise')
MediumVioletRed = Color('#C71585', name='MediumVioletRed')
MidnightBlue = Color('#191970', name='MidnightBlue')
MintCream = Color('#F5FFFA', name='MintCream')
MistyRose = Color('#FFE4E1', name='MistyRose')
Moccasin = Color('#FFE4B5', name='Moccasin')
NavajoWhite = Color('#FFDEAD', name='NavajoWhite')
Navy = Color('#000080', name='Navy')
OldLace = Color('#FDF5E6', name='OldLace')
Olive = Color('#808000', name='Olive')
OliveDrab = Color('#6B8E23', name='OliveDrab')
Orange = Color('#FFA500', name='Orange')
OrangeRed = Color('#FF4500', name='OrangeRed')
Orchid = Color('#DA70D6', name='Orchid')
PaleGoldenRod = Color('#EEE8AA', name='PaleGoldenRod')
PaleGreen = Color('#98FB98', name='PaleGreen')
PaleTurquoise = Color('#AFEEEE', name='PaleTurquoise')
PaleVioletRed = Color('#DB7093', name='PaleVioletRed')
PapayaWhip = Color('#FFEFD5', name='PapayaWhip')
PeachPuff = Color('#FFDAB9', name='PeachPuff')
Peru = Color('#CD853F', name='Peru')
Pink = Color('#FFC0CB', name='Pink')
Plum = Color('#DDA0DD', name='Plum')
PowderBlue = Color('#B0E0E6', name='PowderBlue')
Purple = Color('#800080', name='Purple')
RebeccaPurple = Color('#663399', name='RebeccaPurple')
Red = Color('#FF0000', name='Red')
RosyBrown = Color('#BC8F8F', name='RosyBrown')
RoyalBlue = Color('#4169E1', name='RoyalBlue')
SaddleBrown = Color('#8B4513', name='SaddleBrown')
Salmon = Color('#FA8072', name='Salmon')
SandyBrown = Color('#F4A460', name='SandyBrown')
SeaGreen = Color('#2E8B57', name='SeaGreen')
SeaShell = Color('#FFF5EE', name='SeaShell')
Sienna = Color('#A0522D', name='Sienna')
Silver = Color('#C0C0C0', name='Silver')
SkyBlue = Color('#87CEEB', name='SkyBlue')
SlateBlue = Color('#6A5ACD', name='SlateBlue')
SlateGray = Color('#708090', name='SlateGray')
SlateGrey = Color('#708090', name='SlateGrey')
Snow = Color('#FFFAFA', name='Snow')
SpringGreen = Color('#00FF7F', name='SpringGreen')
SteelBlue = Color('#4682B4', name='SteelBlue')
Tan = Color('#D2B48C', name='Tan')
Teal = Color('#008080', name='Teal')
Thistle = Color('#D8BFD8', name='Thistle')
Tomato = Color('#FF6347', name='Tomato')
Turquoise = Color('#40E0D0', name='Turquoise')
Violet = Color('#EE82EE', name='Violet')
Wheat = Color('#F5DEB3', name='Wheat')
White = Color('#FFFFFF', name='White')
WhiteSmoke = Color('#F5F5F5', name='WhiteSmoke')
Yellow = Color('#FFFF00', name='Yellow')
YellowGreen = Color('#9ACD32', name='YellowGreen')

R = Color('#FF0000', name='R')  # red
G = Color('#008000', name='G')  # green
B = Color('#0000FF', name='B')  # blue

C = Color('#00FFFF', name='C')  # cyan
M = Color('#FF00FF', name='M')  # magenta
Y = Color('#FFFF00', name='Y')  # yellow

T = Color('#0000', name='T')  # transparent
K = Color('#000000', name='K')  # black
W = Color('#FFFFFF', name='W')  # white
A = Color('#808080', name='A')  # gray
O = Color('#FFA500', name='O')  # orange

# define palettes
Pero = Palette(('#1047b9', '#328c00', '#f19000', '#4cc7c5', '#8f8f15', '#edbb00', '#786dff', '#b34e00', '#80bfbd', '#c88812', '#c5ca3d', '#7bb6ff', '#45438a', '#188183', '#838183', '#457ec6', '#7f2200', '#4c4e4c'), name='Pero')
Accent = Palette(('#7fc97f', '#beaed4', '#fdc086', '#ffff99', '#386cb0', '#f0027f', '#bf5b17', '#666666'), name='Accent')
Dark = Palette(('#1b9e77', '#d95f02', '#7570b3', '#e7298a', '#66a61e', '#e6ab02', '#a6761d', '#666666'), name='Dark')
Paired = Palette(('#a6cee3', '#1f78b4', '#b2df8a', '#33a02c', '#fb9a99', '#e31a1c', '#fdbf6f', '#ff7f00', '#cab2d6', '#6a3d9a', '#ffff99', '#b15928'), name='Paired')
Pastel1 = Palette(('#fbb4ae', '#b3cde3', '#ccebc5', '#decbe4', '#fed9a6', '#ffffcc', '#e5d8bd', '#fddaec', '#f2f2f2'), name='Pastel1')
Pastel2 = Palette(('#b3e2cd', '#fdcdac', '#cbd5e8', '#f4cae4', '#e6f5c9', '#fff2ae', '#f1e2cc', '#cccccc'), name='Pastel2')
Set1 = Palette(('#e41a1c', '#377eb8', '#4daf4a', '#984ea3', '#ff7f00', '#ffff33', '#a65628', '#f781bf', '#999999'), name='Set1')
Set2 = Palette(('#66c2a5', '#fc8d62', '#8da0cb', '#e78ac3', '#a6d854', '#ffd92f', '#e5c494', '#b3b3b3'), name='Set2')
Set3 = Palette(('#8dd3c7', '#ffffb3', '#bebada', '#fb8072', '#80b1d3', '#fdb462', '#b3de69', '#fccde5', '#d9d9d9', '#bc80bd', '#ccebc5', '#ffed6f'), name='Set3')
Spectral = Palette(("#3288bd", "#66c2a5", "#abdda4", "#e6f598", "#ffffbf", "#fee08b", "#fdae61", "#f46d43", "#d53e4f"), name='Spectral')

Blues = Palette(('#f7fbff', '#deebf7', '#c6dbef', '#9ecae1', '#6baed6', '#4292c6', '#2171b5', '#08519c', '#08306b'), name='Blues')
Greens = Palette(('#f7fcf5', '#e5f5e0', '#c7e9c0', '#a1d99b', '#74c476', '#41ab5d', '#238b45', '#006d2c', '#00441b'), name='Greens')
Greys = Palette(('#ffffff', '#f0f0f0', '#d9d9d9', '#bdbdbd', '#969696', '#737373', '#525252', '#252525', '#000000'), name='Greys')
Oranges = Palette(('#fff5eb', '#fee6ce', '#fdd0a2', '#fdae6b', '#fd8d3c', '#f16913', '#d94801', '#a63603', '#7f2704'), name='Oranges')
Purples = Palette(('#fcfbfd', '#efedf5', '#dadaeb', '#bcbddc', '#9e9ac8', '#807dba', '#6a51a3', '#54278f', '#3f007d'), name='Purples')
Reds = Palette(('#fff5f0', '#fee0d2', '#fcbba1', '#fc9272', '#fb6a4a', '#ef3b2c', '#cb181d', '#a50f15', '#67000d'), name='Reds')

Inferno = Palette(('#fcfea4', '#f8c931', '#f98c09', '#e35832', '#ba3655', '#88216a', '#550f6d', '#1f0c47', '#000003'), name='Inferno')
Magma = Palette(('#fbfcbf', '#fec286', '#fb8660', '#e55063', '#b53679', '#812581', '#4f117b', '#1b1044', '#000003'), name='Magma')
Plasma = Palette(('#eff821', '#fdc328', '#f79341', '#e56b5c', '#ca4678', '#a82296', '#7c02a7', '#4a02a0', '#0c0786'), name='Plasma')
Viridis = Palette(('#fde724', '#aadb32', '#5bc862', '#27ad80', '#208f8c', '#2c718e', '#3b518a', '#472b7a', '#440154'), name='Viridis')

BrBG = Palette(('#8c510a', '#bf812d', '#dfc27d', '#f6e8c3', '#f5f5f5', '#c7eae5', '#80cdc1', '#35978f', '#01665e'), name='BrBG')
BuGn = Palette(('#f7fcfd', '#e5f5f9', '#ccece6', '#99d8c9', '#66c2a4', '#41ae76', '#238b45', '#006d2c', '#00441b'), name='BuGn')
BuPu = Palette(('#f7fcfd', '#e0ecf4', '#bfd3e6', '#9ebcda', '#8c96c6', '#8c6bb1', '#88419d', '#810f7c', '#4d004b'), name='BuPu')
GnBu = Palette(('#f7fcf0', '#e0f3db', '#ccebc5', '#a8ddb5', '#7bccc4', '#4eb3d3', '#2b8cbe', '#0868ac', '#084081'), name='GnBu')
OrRd = Palette(('#fff7ec', '#fee8c8', '#fdd49e', '#fdbb84', '#fc8d59', '#ef6548', '#d7301f', '#b30000', '#7f0000'), name='OrRd')
PiYG = Palette(('#c51b7d', '#de77ae', '#f1b6da', '#fde0ef', '#f7f7f7', '#e6f5d0', '#b8e186', '#7fbc41', '#4d9221'), name='PiYG')
PRGn = Palette(('#762a83', '#9970ab', '#c2a5cf', '#e7d4e8', '#f7f7f7', '#d9f0d3', '#a6dba0', '#5aae61', '#1b7837'), name='PRGn')
PuBu = Palette(('#fff7fb', '#ece7f2', '#d0d1e6', '#a6bddb', '#74a9cf', '#3690c0', '#0570b0', '#045a8d', '#023858'), name='PuBu')
PuBuGn = Palette(('#fff7fb', '#ece2f0', '#d0d1e6', '#a6bddb', '#67a9cf', '#3690c0', '#02818a', '#016c59', '#014636'), name='PuBuGn')
PuOr = Palette(('#542788', '#8073ac', '#b2abd2', '#d8daeb', '#f7f7f7', '#fee0b6', '#fdb863', '#e08214', '#b35806'), name='PuOr')
PuRd = Palette(('#f7f4f9', '#e7e1ef', '#d4b9da', '#c994c7', '#df65b0', '#e7298a', '#ce1256', '#980043', '#67001f'), name='PuRd')
RdBu = Palette(('#b2182b', '#d6604d', '#f4a582', '#fddbc7', '#f7f7f7', '#d1e5f0', '#92c5de', '#4393c3', '#2166ac'), name='RdBu')
RdGy = Palette(('#b2182b', '#d6604d', '#f4a582', '#fddbc7', '#ffffff', '#e0e0e0', '#bababa', '#878787', '#4d4d4d'), name='RdGy')
RdPu = Palette(('#fff7f3', '#fde0dd', '#fcc5c0', '#fa9fb5', '#f768a1', '#dd3497', '#ae017e', '#7a0177', '#49006a'), name='RdPu')
RdYlBu = Palette(('#d73027', '#f46d43', '#fdae61', '#fee090', '#ffffbf', '#e0f3f8', '#abd9e9', '#74add1', '#4575b4'), name='RdYlBu')
RdYlGn = Palette(('#d73027', '#f46d43', '#fdae61', '#fee08b', '#ffffbf', '#d9ef8b', '#a6d96a', '#66bd63', '#1a9850'), name='RdYlGn')
YlGn = Palette(('#ffffe5', '#f7fcb9', '#d9f0a3', '#addd8e', '#78c679', '#41ab5d', '#238443', '#006837', '#004529'), name='YlGn')
YlGnBu = Palette(('#ffffd9', '#edf8b1', '#c7e9b4', '#7fcdbb', '#41b6c4', '#1d91c0', '#225ea8', '#253494', '#081d58'), name='YlGnBu')
YlOrBr = Palette(('#ffffe5', '#fff7bc', '#fee391', '#fec44f', '#fe9929', '#ec7014', '#cc4c02', '#993404', '#662506'), name='YlOrBr')
YlOrRd = Palette(('#ffffcc', '#ffeda0', '#fed976', '#feb24c', '#fd8d3c', '#fc4e2a', '#e31a1c', '#bd0026', '#800026'), name='YlOrRd')

# define gradients
palettes = (Spectral,
    Blues, Greens, Greys, Oranges, Purples, Reds,
    Inferno, Magma, Plasma, Viridis,
    BrBG, BuGn, BuPu, GnBu, OrRd, PiYG, PRGn, PuBu, PuBuGn, PuOr, PuRd, RdBu, RdGy, RdPu, RdYlBu, RdYlGn, YlGn, YlGnBu, YlOrBr, YlOrRd)

for palette in palettes:
    Gradient(palette, name=palette.name)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import math
from .. enums import *
from .. properties import *
from . formatter import Formatter


class ScalarFormatter(Formatter):
    """
    This formatter tool uses an automatic formatting of given number according
    to current 'domain', 'precision' and additional settings. It is typically
    used to format labels for standard numeric axes.
    
    A scientific notation style (e.g. 2e6) can be used to format numbers
    exceeding specific threshold, which can be specified by the 'sci_threshold'
    property as a power of 10. (E.g. if set to 3, all the values grater than
    1000 or smaller than -1000 wil be formatted with scientific notation.)
    
    The scientific notation suffix can be removed from labels by setting the
    'hide_suffix' property to True and it becomes available using the 'suffix'
    method. Specific formatting of the suffix can also be defined by the
    'suffix_template' property expecting the power as input (e.g. 'e{:.0f}').
    
    Properties:
        
        sci_notation: bool
            Specifies whether the scientific notation (e.g. 2e6) is enabled
            (True) or disabled (False).
        
        sci_threshold: int
            Specifies the threshold for 'domain' from which the scientific
            notation should be used as a power of 10.
        
        hide_suffix: bool
            Specifies whether the scientific notation suffix should be removed
            from the formatted value (True). The actual suffix can be retrieved
            by the 'suffix' method.
        
        suffix_template: str, None or UNDEF
            Specifies the format()-style template to be used for suffix
            formatting expecting the power as input (e.g. 'e{:.0f}').
    """
    
    sci_notation = BoolProperty(True, dynamic=False)
    sci_threshold = IntProperty(5, dynamic=False)
    
    hide_suffix = BoolProperty(False, dynamic=False)
    suffix_template = StringProperty(" (e{:.0f})", dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of ScalarFormatter."""
        
        super().__init__(**overrides)
        
        # init buffers
        self._power = 0
        self._last_digit = 0
        self._has_suffix = False
        self._is_dirty = True
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_scalar_formatter_property_changed)
    
    
    def format(self, value, *args, **kwargs):
        """
        Formats a given value using scalar formatting.
        
        Args:
            value: float
                Value to be formatted.
        
        Returns:
            str
                Formatted label.
        """
        
        # init formatting
        if self._is_dirty:
            self._init_formatting()
        
        # use suffix
        if self.hide_suffix and self._has_suffix:
            value = value / (10.**self._power)
            template = "{:1.%df}" % abs(self._last_digit - self._power)
        
        # use scientific notation
        elif self.sci_notation and self._has_suffix:
            
            last_digit = 0
            if value:
                value_power = int(math.floor(math.log10(abs(value))))
                if value_power > self._last_digit:
                    last_digit = self._last_digit - value_power
            
            template = "{:1.%de}" % abs(last_digit)
        
        # use current precision
        elif self._last_digit < 0:
            template = "{:0.%df}" % abs(self._last_digit)
        
        # make integer
        else:
            template = "{:0.0f}"
        
        # apply format
        label = template.format(value)
        
        # avoid '-0'
        if label[0] == '-' and float(label) == 0:
            label = label[1:]
        
        return label
    
    
    def suffix(self, *args, **kwargs):
        """
        Gets current labels suffix (e.g. e-2).
        
        Returns:
            str
                Labels suffix.
        """
        
        # init formatting
        if self._is_dirty:
            self._init_formatting()
        
        # return suffix
        if self.hide_suffix and self._has_suffix and self.suffix_template:
            return self.suffix_template.format(self._power)
        
        # not defined
        return ""
    
    
    def _init_formatting(self):
        """Initializes formatting based on current range."""
        
        # get range and precision
        domain = self.domain or 1.
        precision = self.precision or 1.
        
        # calc power and digits
        self._power = int(math.floor(math.log10(abs(domain))))
        self._has_suffix = abs(self._power) >= abs(self.sci_threshold)
        self._last_digit = int(math.floor(math.log10(abs(precision))))
        
        # reset flag
        self._is_dirty = False
    
    
    def _on_scalar_formatter_property_changed(self, evt=None):
        """Called after a property has changed."""
        
        self._is_dirty = True



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. properties import *


class Graphics(PropertySet):
    """
    Abstract base class for all graphical objects.
    
    Properties:
        
        tag: str, callable, None or UNDEF
            Specifies the unique tag of the object. If not provided on a class
            initialization it is set automatically using object id. The value is
            often used as id when drawn to SVG or as object identifier in
            complex graphics like charts.
        
        visible: bool, callable
            Specifies whether the object should be drawn or skipped.
        
        z_index: int, float or UNDEF
            Specifies the object z-index. The value may be used to provide
            specific order for objects drawing in complex graphics like charts.
            Bigger the value is, more on top the graphics should be drawn. Any
            specific use of this must be implemented by a user to draw graphics
            in desired order.
    """
    
    tag = StringProperty(UNDEF, nullable=True)
    visible = BoolProperty(True)
    z_index = NumProperty(UNDEF)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of Graphics."""
        
        super().__init__(**overrides)
        
        # init default tag
        if self.tag is UNDEF:
            self.tag = "tag_%s" % str(id(self))
    
    
    def show(self, title=None, width=None, height=None, backend=None, **options):
        """
        Shows current graphics in available viewer app. This method makes sure
        appropriate backend canvas is created and provided to the 'draw' method.
        
        Note that is just a convenient scripting shortcut and this method cannot
        be used if the graphics is already part of any UI app.
        
        Args:
            title: str or None
                Viewer frame title.
            
            width: float or None
                Viewer width in device units.
            
            height: float or None
                Viewer height in device units.
            
            backend: pero.BACKEND
                Specific backend to be used. The value must be an item from the
                pero.BACKEND enum.
            
            options: str:any pairs
                Additional parameters for specific backend.
        """
        
        # try get size
        if width is None and self.has_property('width'):
            width = self.get_property('width')
        if height is None and self.has_property('height'):
            height = self.get_property('height')
        
        from .. import backends
        backends.show(self, title, width, height, backend, **options)
    
    
    def export(self, path, width=None, height=None, backend=None, **options):
        """
        Draws current graphics into specified image file using the format
        determined automatically from the file extension. This method makes sure
        appropriate backend canvas is created and provided to the 'draw' method.
        
        Note that is just a convenient scripting shortcut and this method cannot
        be used if the graphics is already part of any UI app.
        
        Args:
            path: str
                Full path of a file to save the image into.
            
            width: float or None
                Image width in device units.
            
            height: float or None
                Image height in device units.
            
            backend: pero.BACKEND
                Specific backend to be used. The value must be an item from the
                pero.BACKEND enum.
            
            options: str:any pairs
                Additional parameters for specific backend.
        """
        
        # try get size
        if width is None and self.has_property('width'):
            width = self.get_property('width')
        if height is None and self.has_property('height'):
            height = self.get_property('height')
        
        from .. import backends
        backends.export(self, path, width, height, backend, **options)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """
        Uses given canvas to draw the graphics.
        
        Args:
            canvas: pero.Canvas
                Canvas to be used for rendering.
            
            source: any
                Data source to be used for calculating callable properties.
            
            overrides: str:any pairs
                Specific properties to be overwritten.
        """
        
        raise NotImplementedError("The 'draw' method is not implemented for '%s'." % self.__class__.__name__)
    
    
    def draw_many(self, canvas, source, **overrides):
        """
        Uses given canvas to draw the graphics for each item in given source.
        In fact this is just a convenient shortcut to call the 'draw' method in
        a loop.
        
        Args:
            canvas: pero.Canvas
                Canvas to be used for rendering.
            
            source: (any,)
                Collection of data source items to be used for calculating
                callable properties.
            
            overrides: str:any pairs
                Specific properties to be overwritten.
        """
        
        # check source
        if source is None or source is UNDEF:
            return
        
        # draw items
        for item in source:
            self.draw(canvas, item, **overrides)
    
    
    def is_visible(self, source=UNDEF, overrides=None):
        """
        Returns True if current graphics is visible.
        
        Args:
            source: any
                Data source to be used for calculating callable properties.
            
            overrides: dict
                Specific properties to be overwritten.
        
        Returns:
            bool
                Returns True if current graphics is visible.
        """
        
        return self.get_property('visible', source, overrides)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from .. drawing import Matrix, Path
from . glyph import Glyph
from . heads import HeadProperty


class Arrow(Glyph):
    """
    Abstract base class for various types of arrow glyphs. An arrow is drawn as
    a simple strait or curved line having optional heads at each side.
    
    When the 'source' argument is provided for drawing it is also submitted to
    the drawing of the heads, together with relevant overrides so it will be
    used for any dynamic property of the head. Note that the 'x', 'y' and
    'angle' properties of the heads are determined by the arrow itself and
    cannot be overwritten.
    
    Properties:
        
        start_head: pero.Head, pero.HEAD, callable, None or UNDEF
            Specifies the head glyph to be drawn at the beginning or the arrow.
            The value can be specified by any item from the pero.HEAD enum or
            pero.Head instance.
        
        end_head: pero.Head, pero.HEAD, callable, None or UNDEF
            Specifies the head glyph to be drawn at the end or the arrow. The
            value can be specified by any item from the pero.HEAD enum or
            pero.Head instance.
        
        line properties:
            Includes pero.LineProperties to specify the arrow line.
        
        fill properties:
            Includes pero.FillProperties to specify the arrow fill.
    """
    
    TYPE = None
    
    start_head = HeadProperty(UNDEF, nullable=True)
    end_head = HeadProperty(UNDEF, nullable=True)
    
    line = Include(LineProperties, line_color=UNDEF)
    fill = Include(FillProperties, fill_color=UNDEF)
    
    
    def draw_heads(self, canvas, source=UNDEF, **overrides):
        """Draws arrow heads into given canvas."""
        
        # get properties
        start_head = self.get_property('start_head', source, overrides)
        end_head = self.get_property('end_head', source, overrides)
        
        # draw start head
        if start_head:
            
            canvas.set_pen_by(self, source=source, overrides=overrides)
            canvas.set_brush_by(self, source=source, overrides=overrides)
            
            head_overrides = self.get_child_overrides('start_head', overrides)
            start_head.draw(canvas, source=source, **head_overrides)
        
        # draw end head
        if end_head:
            
            canvas.set_pen_by(self, source=source, overrides=overrides)
            canvas.set_brush_by(self, source=source, overrides=overrides)
            
            head_overrides = self.get_child_overrides('end_head', overrides)
            end_head.draw(canvas, source=source, **head_overrides)
    
    
    @staticmethod
    def create(symbol, **overrides):
        """
        Initializes a new instance of arrow according to given definition. The
        definition consists of the arrow type symbol surrounded by optional
        heads at both ends (e.g.'<->'). The arrow symbol must be an item from
        the pero.ARROW enum and it is required. The optional symbols for
        individual heads can be specified as an item from the pero.HEAD enum.
        
        Args:
            symbol: str
                Arrow definition.
        
        Returns:
            pero.Arrow
                Initialized arrow.
        """
        
        # get arrow type
        arrow_type = None
        for arrow in ARROW:
            if arrow in symbol:
                arrow_type = arrow
                break
        
        if arrow_type is None:
            raise ValueError("Unknown arrow style! -> '%s'" % symbol)
        
        # split symbol into heads
        start_head, end_head = symbol.split(arrow_type)
        
        # set heads
        if 'start_head' not in overrides:
            if start_head and start_head not in HEAD:
                raise ValueError("Unknown start head! -> '%s'" % symbol)
            overrides['start_head'] = start_head or None
        
        if 'end_head' not in overrides:
            if end_head and end_head not in HEAD:
                raise ValueError("Unknown end head! -> '%s'" % symbol)
            overrides['end_head'] = end_head or None
        
        # init arrow
        if arrow_type == ARROW_ARC:
            return ArcArrow(**overrides)
        
        elif arrow_type == ARROW_CONNECT_LINE:
            return ConnectorArrow(**overrides)
        
        elif arrow_type == ARROW_CONNECT_CURVE:
            if 'curve' not in overrides:
                overrides['curve'] = .85
            return ConnectorArrow(**overrides)
        
        elif arrow_type == ARROW_CURVE:
            return CurveArrow(**overrides)
        
        elif arrow_type == ARROW_LINE:
            return LineArrow(**overrides)
        
        elif arrow_type == ARROW_BOW:
            return BowArrow(**overrides)
        
        elif arrow_type == ARROW_RAY:
            return RayArrow(**overrides)


class ArcArrow(Arrow):
    """
    This type of arrow is drawn as a simple circular arc defined by its center
    coordinates, start and end angles and the drawing direction.
    
    Properties:
        
        x: int, float, callable
            Specifies the x-coordinate of the arc center.
        
        y: int, float, callable
            Specifies the y-coordinate of the arc center.
        
        radius: int, float, callable
            Specifies the arc radius.
        
        clockwise: bool, callable
            Specifies the drawing direction. If set to True the arc is drawn
            clockwise, otherwise anti-clockwise.
        
        start_angle properties:
            Includes pero.AngleProperties to specify the start angle.
        
        end_angle properties:
            Includes pero.AngleProperties to specify the end angle.
    """
    
    TYPE = ARROW_ARC
    
    x = NumProperty(0)
    y = NumProperty(0)
    radius = NumProperty(0)
    clockwise = BoolProperty(True)
    start_angle = Include(AngleProperties, prefix="start_")
    end_angle = Include(AngleProperties, prefix="end_")
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw arrow."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        start_angle = AngleProperties.get_angle(self, 'start_', ANGLE_RAD, source, overrides)
        end_angle = AngleProperties.get_angle(self, 'end_', ANGLE_RAD, source, overrides)
        radius = self.get_property('radius', source, overrides)
        clockwise = self.get_property('clockwise', source, overrides)
        
        # make path
        path = Path()
        path.arc(x, y, radius, start_angle, end_angle, clockwise)
        
        # get coordinates
        x1 = x + radius * math.cos(start_angle)
        y1 = y + radius * math.sin(start_angle)
        x2 = x + radius * math.cos(end_angle)
        y2 = y + radius * math.sin(end_angle)
        
        direction = 1 if clockwise else -1
        start_angle = start_angle - 0.5*math.pi * direction
        end_angle = end_angle + 0.5*math.pi * direction
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # start drawing group
        canvas.group(tag, "arrow")
        
        # draw path
        canvas.draw_path(path)
        
        # draw heads
        self.draw_heads(canvas,
            source = source,
            start_head_x = x1,
            start_head_y = y1,
            start_head_angle = start_angle,
            start_head_angle_units = ANGLE_RAD,
            end_head_x = x2,
            end_head_y = y2,
            end_head_angle = end_angle,
            end_head_angle_units = ANGLE_RAD,
            **overrides)
        
        # end drawing group
        canvas.ungroup()


class BowArrow(Arrow):
    """
    This type of arrow is drawn as a simple circular arc defined by its start
    and end coordinates, expected radius and the drawing direction.
    
    Properties:
        
        x1: int, float, callable
            Specifies the x-coordinate of the arrow start.
        
        y1: int, float, callable
            Specifies the y-coordinate of the arrow start.
        
        x2: int, float, callable
            Specifies the x-coordinate of the arrow end.
        
        y2: int, float, callable
            Specifies the y-coordinate of the arrow end.
        
        radius: int, float, callable
            Specifies the arc radius.
        
        large: bool
            Specifies which of the possible arcs will be drawn.
        
        clockwise: bool, callable
            Specifies the drawing direction. If set to True the arc is drawn
            clockwise, otherwise anti-clockwise.
    """
    
    TYPE = ARROW_BOW
    
    x1 = NumProperty(0)
    y1 = NumProperty(0)
    x2 = NumProperty(0)
    y2 = NumProperty(0)
    radius = NumProperty(0)
    large = BoolProperty(False)
    clockwise = BoolProperty(True)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw arrow."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x1 = self.get_property('x1', source, overrides)
        y1 = self.get_property('y1', source, overrides)
        x2 = self.get_property('x2', source, overrides)
        y2 = self.get_property('y2', source, overrides)
        radius = self.get_property('radius', source, overrides)
        large = self.get_property('large', source, overrides)
        clockwise = self.get_property('clockwise', source, overrides)
        
        # make path
        path = Path()
        path.move_to(x1, y1)
        path.bow_to(x2, y2, radius, large, clockwise)
        
        # get edge angles
        start_angle = path.start_angle() - math.pi
        end_angle = path.end_angle()
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # start drawing group
        canvas.group(tag, "arrow")
        
        # draw path
        canvas.draw_path(path)
        
        # draw heads
        self.draw_heads(canvas,
            source = source,
            start_head_x = x1,
            start_head_y = y1,
            start_head_angle = start_angle,
            start_head_angle_units = ANGLE_RAD,
            end_head_x = x2,
            end_head_y = y2,
            end_head_angle = end_angle,
            end_head_angle_units = ANGLE_RAD,
            **overrides)
        
        # end drawing group
        canvas.ungroup()


class ConnectorArrow(Arrow):
    """
    This type of arrow is typically used to draw connections in node diagrams.
    It is drawn as a stepped-line or curve between two end points, with the
    heads facing left/right or up/down according to the 'orientation' property.
    
    Properties:
        
        x1: int, float, callable
            Specifies the x-coordinate of the arrow start.
        
        y1: int, float, callable
            Specifies the y-coordinate of the arrow start.
        
        x2: int, float, callable
            Specifies the x-coordinate of the arrow end.
        
        y2: int, float, callable
            Specifies the y-coordinate of the arrow end.
        
        orientation: pero.ORIENTATION, callable
            Specifies the main orientation of the line and heads as any item
            from the pero.ORIENTATION enum.
        
        pivot: float, callable
            Specifies the relative position of the connection line pivot point
            as a value between 0 and 1, where 0 means the arrow start, while 1
            means the arrow end. If set to 0 or 1 exactly, the arrow changes to
            simple L-shape line instead of Z-shape line.
        
        curve: float, callable
            Specifies the force used to curve the connection line as a value
            between 0 and 1, where 0 means strait line, while 1 means
            fully curved line.
    """
    
    TYPE = ARROW_CONNECT_LINE
    
    x1 = NumProperty(0)
    y1 = NumProperty(0)
    x2 = NumProperty(0)
    y2 = NumProperty(0)
    
    orientation = EnumProperty(ORI_HORIZONTAL, enum=ORIENTATION)
    pivot = RangeProperty(0.5, minimum=0, maximum=1)
    curve = RangeProperty(0, minimum=0, maximum=1)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw arrow."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x1 = self.get_property('x1', source, overrides)
        y1 = self.get_property('y1', source, overrides)
        x2 = self.get_property('x2', source, overrides)
        y2 = self.get_property('y2', source, overrides)
        orientation = self.get_property('orientation', source, overrides)
        pivot = self.get_property('pivot', source, overrides)
        curve = self.get_property('curve', source, overrides)
        
        # get curvature
        curvature = 2*curve if curve else 1
        
        # get coords
        if orientation == ORI_HORIZONTAL:
            cx1 = x1 + min(1, curvature*pivot)*(x2-x1)
            cy1 = y1
            cx2 = x2 - min(1, curvature*(1-pivot))*(x2-x1)
            cy2 = y2
        
        else:
            cx1 = x1
            cy1 = y1 + min(1, curvature*pivot)*(y2-y1)
            cx2 = x2
            cy2 = y2 - min(1, curvature*(1-pivot))*(y2-y1)
        
        # get angles
        if x1 == cx1 and y1 == cy1:
            start_angle = math.atan2(y1-cy2, x1-cx2)
            end_angle = math.atan2(y2-cy2, x2-cx2)
        
        elif x2 == cx2 and y2 == cy2:
            start_angle = math.atan2(y1-cy1, x1-cx1)
            end_angle = math.atan2(y2-cy1, x2-cx1)
        
        else:
            start_angle = math.atan2(y1-cy1, x1-cx1)
            end_angle = math.atan2(y2-cy2, x2-cx2)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # start drawing group
        canvas.group(tag, "arrow")
        
        # draw line
        if curve == 0:
            canvas.draw_lines(((x1, y1), (cx1, cy1), (cx2, cy2), (x2, y2)))
        
        # draw curve
        else:
            path = Path()
            path.move_to(x1, y1)
            path.curve_to(cx1, cy1, cx2, cy2, x2, y2)
            canvas.draw_path(path)
        
        # draw heads
        self.draw_heads(canvas,
            source = source,
            start_head_x = x1,
            start_head_y = y1,
            start_head_angle = start_angle,
            start_head_angle_units = ANGLE_RAD,
            end_head_x = x2,
            end_head_y = y2,
            end_head_angle = end_angle,
            end_head_angle_units = ANGLE_RAD,
            **overrides)
        
        # end drawing group
        canvas.ungroup()


class CurveArrow(Arrow):
    """
    This type of arrow is drawn as a Bezier curve between two anchor points.
    Unlike for normal curve, the control points are defined by relative position
    between anchor points.
    
    Properties:
        
        x1: int, float, callable
            Specifies the x-coordinate of the arrow start.
        
        y1: int, float, callable
            Specifies the y-coordinate of the arrow start.
        
        x2: int, float, callable
            Specifies the x-coordinate of the arrow end.
        
        y2: int, float, callable
            Specifies the y-coordinate of the arrow end.
        
        cx1: int, float, callable
            Specifies the relative x-coordinate of the arrow start control point
            as a value between 0 and 1, where 0 means the arrow start, while 1
            means the arrow end.
        
        cy1: int, float, callable
            Specifies the relative y-coordinate of the arrow start control point
            as a value between 0 and 1, where 0 means the arrow start, while 1
            means the arrow end.
        
        cx2: int, float, callable
            Specifies the relative x-coordinate of the arrow end control point
            as a value between 0 and 1, where 0 means the arrow end, while 1
            means the arrow start.
        
        cy2: int, float, callable
            Specifies the relative y-coordinate of the arrow end control point
            as a value between 0 and 1, where 0 means the arrow end, while 1
            means the arrow start.
    """
    
    TYPE = ARROW_CURVE
    
    x1 = NumProperty(0)
    y1 = NumProperty(0)
    x2 = NumProperty(0)
    y2 = NumProperty(0)
    
    cx1 = NumProperty(0.85)
    cy1 = NumProperty(0.25)
    cx2 = NumProperty(0.85)
    cy2 = NumProperty(0.25)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw arrow."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x1 = self.get_property('x1', source, overrides)
        y1 = self.get_property('y1', source, overrides)
        x2 = self.get_property('x2', source, overrides)
        y2 = self.get_property('y2', source, overrides)
        cx1 = self.get_property('cx1', source, overrides)
        cy1 = self.get_property('cy1', source, overrides)
        cx2 = self.get_property('cx2', source, overrides)
        cy2 = self.get_property('cy2', source, overrides)
        
        # calc absolute coords
        cx1 = x1 + cx1*(x2 - x1)
        cx2 = x2 - cx2*(x2 - x1)
        cy1 = y1 + cy1*(y2 - y1)
        cy2 = y2 - cy2*(y2 - y1)
        
        # make path
        path = Path()
        path.move_to(x1, y1)
        path.curve_to(cx1, cy1, cx2, cy2, x2, y2)
        
        # get angles
        start_angle = path.start_angle() + math.pi
        end_angle = path.end_angle()
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # start drawing group
        canvas.group(tag, "arrow")
        
        # draw path
        canvas.draw_path(path)
        
        # draw heads
        self.draw_heads(canvas,
            source = source,
            start_head_x = x1,
            start_head_y = y1,
            start_head_angle = start_angle,
            start_head_angle_units = ANGLE_RAD,
            end_head_x = x2,
            end_head_y = y2,
            end_head_angle = end_angle,
            end_head_angle_units = ANGLE_RAD,
            **overrides)
        
        # end drawing group
        canvas.ungroup()


class LineArrow(Arrow):
    """
    This type of arrow is drawn as a strait line between two end points.
    
    Properties:
        
        x1: int, float, callable
            Specifies the x-coordinate of the arrow start.
        
        y1: int, float, callable
            Specifies the y-coordinate of the arrow start.
        
        x2: int, float, callable
            Specifies the x-coordinate of the arrow end.
        
        y2: int, float, callable
            Specifies the y-coordinate of the arrow end.
    """
    
    TYPE = ARROW_LINE
    
    x1 = NumProperty(0)
    y1 = NumProperty(0)
    x2 = NumProperty(0)
    y2 = NumProperty(0)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw the glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x1 = self.get_property('x1', source, overrides)
        y1 = self.get_property('y1', source, overrides)
        x2 = self.get_property('x2', source, overrides)
        y2 = self.get_property('y2', source, overrides)
        
        # get angle
        angle = math.atan2(y2-y1, x2-x1)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # start drawing group
        canvas.group(tag, "arrow")
        
        # draw line
        canvas.draw_line(x1, y1, x2, y2)
        
        # set brush
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw heads
        self.draw_heads(canvas,
            source = source,
            start_head_x = x1,
            start_head_y = y1,
            start_head_angle = angle+math.pi,
            start_head_angle_units = ANGLE_RAD,
            end_head_x = x2,
            end_head_y = y2,
            end_head_angle = angle,
            end_head_angle_units = ANGLE_RAD,
            **overrides)
        
        # end drawing group
        canvas.ungroup()


class PathArrow(Arrow):
    """
    This type of arrow is drawn as a given path rotated and scaled the way that
    its end points fit into arrow start and end.
    
    Properties:
        
        x1: int, float, callable
            Specifies the x-coordinate of the arrow start.
        
        y1: int, float, callable
            Specifies the y-coordinate of the arrow start.
        
        x2: int, float, callable
            Specifies the x-coordinate of the arrow end.
        
        y2: int, float, callable, None or UNDEF
            Specifies the y-coordinate of the arrow end.
        
        path: pero.Path, callable
            Specifies the path.
    """
    
    x1 = NumProperty(0)
    y1 = NumProperty(0)
    x2 = NumProperty(0)
    y2 = NumProperty(0)
    
    path = Property(UNDEF, types=(Path,), nullable=True)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw arrow."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x1 = self.get_property('x1', source, overrides)
        y1 = self.get_property('y1', source, overrides)
        x2 = self.get_property('x2', source, overrides)
        y2 = self.get_property('y2', source, overrides)
        path = self.get_property('path', source, overrides)
        
        # check data
        if not path:
            return
        
        # get path endpoints
        p1x, p1y = path.start()
        p2x, p2y = path.end()
        path_length = math.sqrt((p2x-p1x)**2 + (p2y-p1y)**2)
        
        # init matrix
        matrix = Matrix()
        
        # strait path
        angle = math.atan2(p2y-p1y, p2x-p1x)
        matrix.translate(-p1x, -p1y)
        matrix.rotate(-angle)
        
        # scale to final arrow
        length = math.sqrt((x2-x1)**2 + (y2-y1)**2)
        scale = length/path_length
        matrix.scale(scale, scale)
        
        # rotate to final arrow
        angle = math.atan2(y2-y1, x2-x1)
        matrix.rotate(angle)
        
        # move to final position
        matrix.translate(x1, y1)
        
        # apply to path
        path = path.transformed(matrix)
        
        # get angles
        start_angle = path.start_angle() + math.pi
        end_angle = path.end_angle()
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # start drawing group
        canvas.group(tag, "arrow")
        
        # draw path
        canvas.draw_path(path)
        
        # draw heads
        self.draw_heads(canvas,
            source = source,
            start_head_x = x1,
            start_head_y = y1,
            start_head_angle = start_angle,
            start_head_angle_units = ANGLE_RAD,
            end_head_x = x2,
            end_head_y = y2,
            end_head_angle = end_angle,
            end_head_angle_units = ANGLE_RAD,
            **overrides)
        
        # end drawing group
        canvas.ungroup()


class RayArrow(Arrow):
    """
    This type of arrow is drawn as a strait line with specified origin, length
    and angle.
    
    Properties:
        
        x: int, float, callable
            Specifies the x-coordinate of the arrow origin.
        
        y: int, float, callable
            Specifies the y-coordinate of the arrow origin.
        
        length: int, float, callable
            Specifies the full arrow length.
        
        angle properties:
            Includes pero.AngleProperties to specify the line angle.
    """
    
    TYPE = ARROW_RAY
    
    x = NumProperty(0)
    y = NumProperty(0)
    length = NumProperty(0)
    angle = Include(AngleProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw the glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x1 = self.get_property('x', source, overrides)
        y1 = self.get_property('y', source, overrides)
        length = self.get_property('length', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # calc end point
        x2 = x1 + length * math.cos(angle)
        y2 = y1 + length * math.sin(angle)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # start drawing group
        canvas.group(tag, "arrow")
        
        # draw line
        canvas.draw_line(x1, y1, x2, y2)
        
        # draw heads
        self.draw_heads(canvas,
            source = source,
            start_head_x = x1,
            start_head_y = y1,
            start_head_angle = angle+math.pi,
            start_head_angle_units = ANGLE_RAD,
            end_head_x = x2,
            end_head_y = y2,
            end_head_angle = angle,
            end_head_angle_units = ANGLE_RAD,
            **overrides)
        
        # end drawing group
        canvas.ungroup()


class ArrowProperty(Property):
    """
    Defines an arrow property, which simplifies an arrow definition by
    converting specific expression into an instance of corresponding arrow
    glyph.
    
    The expression consists of the arrow type symbol surrounded by optional
    heads at both ends (e.g.'<->'). The arrow symbol must be an item from the
    pero.ARROW enum and it is required. The optional symbols for individual
    heads can be specified as an item from the pero.HEAD enum.
    """
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of ArrowProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (Arrow, str)
        
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # check type
        if isinstance(value, Arrow):
            return value
        
        # parse main
        value = super().parse(value)
        
        # allow UNDEF and None
        if value is UNDEF or value is None:
            return value
        
        # check func
        if callable(value):
            return value
        
        # convert to arrow
        return Arrow.create(value)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . values import *
from . enum import Enum

# define orientation
ORI_HORIZONTAL = HORIZONTAL
ORI_VERTICAL = VERTICAL

ORIENTATION = Enum(
    HORIZONTAL = ORI_HORIZONTAL,
    VERTICAL = ORI_VERTICAL)

# define generic positions
POS_LEFT = LEFT
POS_RIGHT = RIGHT

POS_TOP = TOP
POS_BOTTOM = BOTTOM

POS_CENTER = CENTER
POS_MIDDLE = MIDDLE

POS_START = START
POS_END = END

POS_INSIDE = INSIDE
POS_OUTSIDE = OUTSIDE

POS_N = N
POS_NW = NW
POS_NE = NE
POS_S = S
POS_SW = SW
POS_SE = SE
POS_W = W
POS_E = E
POS_C = C

POSITION = Enum(
    
    LEFT = POS_LEFT,
    RIGHT = POS_RIGHT,
    
    TOP = POS_TOP,
    BOTTOM = POS_BOTTOM,
    
    CENTER = POS_CENTER,
    MIDDLE = POS_MIDDLE,
    
    START = POS_START,
    END = POS_END,
    
    INSIDE = POS_INSIDE,
    OUTSIDE = POS_OUTSIDE,
    
    N = POS_N,
    NW = POS_NW,
    NE = POS_NE,
    S = POS_S,
    SW = POS_SW,
    SE = POS_SE,
    W = POS_W,
    E = POS_E,
    C = POS_C)

# define specific positions
POSITION_LR = Enum(
    LEFT = POS_LEFT,
    RIGHT = POS_RIGHT)

POSITION_LRC = Enum(
    LEFT = POS_LEFT,
    RIGHT = POS_RIGHT,
    CENTER = POS_CENTER)

POSITION_TB = Enum(
    TOP = POS_TOP,
    BOTTOM = POS_BOTTOM)

POSITION_TBC = Enum(
    TOP = POS_TOP,
    BOTTOM = POS_BOTTOM,
    CENTER = POS_CENTER)

POSITION_LRTB = Enum(
    LEFT = POS_LEFT,
    RIGHT = POS_RIGHT,
    TOP = POS_TOP,
    BOTTOM = POS_BOTTOM)

POSITION_LRTBC = Enum(
    LEFT = POS_LEFT,
    RIGHT = POS_RIGHT,
    TOP = POS_TOP,
    BOTTOM = POS_BOTTOM,
    CENTER = POS_CENTER)

POSITION_IOC = Enum(
    INSIDE = POS_INSIDE,
    OUTSIDE = POS_OUTSIDE,
    CENTER = POS_CENTER)

POSITION_SEM = Enum(
    START = POS_START,
    END = POS_END,
    MIDDLE = POS_MIDDLE)

POSITION_TL = Enum(
    TOP = POS_TOP,
    LEFT = POS_LEFT)

POSITION_TR = Enum(
    TOP = POS_TOP,
    RIGHT = POS_RIGHT)

POSITION_BL = Enum(
    BOTTOM = POS_BOTTOM,
    LEFT = POS_LEFT)

POSITION_BR = Enum(
    BOTTOM = POS_BOTTOM,
    RIGHT = POS_RIGHT)

POSITION_COMPASS = Enum(
    N = POS_N,
    NW = POS_NW,
    NE = POS_NE,
    S = POS_S,
    SW = POS_SW,
    SE = POS_SE,
    W = POS_W,
    E = POS_E,
    C = POS_C)

POSITION_COMPASS_LEFT = Enum(
    NW = POS_NW,
    SW = POS_SW,
    W = POS_W)

POSITION_COMPASS_RIGHT = Enum(
    NE = POS_NE,
    SE = POS_SE,
    E = POS_E)

POSITION_COMPASS_TOP = Enum(
    N = POS_N,
    NW = POS_NW,
    NE = POS_NE)

POSITION_COMPASS_BOTTOM = Enum(
    S = POS_S,
    SW = POS_SW,
    SE = POS_SE)

POSITION_COMPASS_CENTER = Enum(
    N = POS_N,
    S = POS_S,
    C = POS_C)

POSITION_COMPASS_MIDDLE = Enum(
    W = POS_W,
    E = POS_E,
    C = POS_C)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . enum import Enum

# define cursors
CURSOR_BLANK = 'blank'
CURSOR_ARROW = 'arrow'
CURSOR_HAND = 'hand'
CURSOR_SIZEWE = 'size_we'
CURSOR_SIZENS = 'size_ns'

CURSOR = Enum(
    BLANK = CURSOR_BLANK,
    ARROW = CURSOR_ARROW,
    HAND = CURSOR_HAND,
    SIZEWE = CURSOR_SIZEWE,
    SIZENS = CURSOR_SIZENS)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import sys
import os.path
from PIL import ImageFont
from .. enums import *


class FontManager(object):
    """
    Font manager holds information about all supported fonts currently
    available. This is mainly used to speed up text size calculations and font
    handling for canvas like SVG, where no specific text size calculation is
    available. For now, only the TrueType or OpenType fonts are supported.
    """
    
    def __init__(self):
        """Initializes a new instance of FontManager."""
        
        self._fonts = {}
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        buff = "["
        for font in self.fonts:
            buff += "\n%s" % font
        buff += "\n]"
        
        return buff
    
    
    def __repr__(self):
        """Gets debug string representation."""
        
        return self.__str__()
    
    
    @property
    def families(self):
        """
        Gets all available font families names.
        
        Returns:
            (str,)
                Tuple of font families names.
        """
        
        return tuple(sorted(self._fonts.keys()))
    
    
    @property
    def fonts(self):
        """
        Gets all available fonts sorted by name.
        
        Returns:
            (pero.Font,)
                Tuple of fonts
        """
        
        fonts = (self._fonts[name] for name in sorted(self._fonts.keys()))
        return tuple(v for f in fonts for v in f)
    
    
    def get_fonts(self, family):
        """
        Gets all fonts from library corresponding to family name.
        
        Args:
            family: str
                Family name of the font.
        
        Returns:
            (pero.Font,)
                Family fonts.
        """
        
        return tuple(self._fonts.get(family, []))
    
    
    def get_font(self, family, style=None, weight=None, loose=True):
        """
        Gets specified font from library.
        
        Args:
            family: str
                Family name of the font.
            
            style: pero.FONT_STYLE or None
                Specifies requested style of the font as any item from the
                pero.FONT_STYLE enum or None if not important.
            
            weight: pero.FONT_WEIGHT or None
                Specifies requested weight of the font as any item from the
                pero.FONT_WEIGHT enum or None if not important.
            
            loose: bool
                If set to True, style and weight might be ignored if not
                available.
        
        Returns:
            pero.Font or None
                Requested font or None if not found.
        """
        
        # get fonts
        fonts = self._fonts.get(family, None)
        if fonts is None:
            return None
        
        # check style
        if style:
            matching = [f for f in fonts if f.style == style]
            if matching or not loose:
                fonts = matching
        
        # check specific weight
        for font in fonts:
            if not weight or font.weight == weight:
                return font
        
        # check light weight loosely
        if weight in FONT_WEIGHTS_LIGHT:
            for font in fonts:
                if font.weight in FONT_WEIGHTS_LIGHT:
                    return font
        
        # check bold weight loosely
        if weight in FONT_WEIGHTS_BOLD:
            for font in fonts:
                if font.weight in FONT_WEIGHTS_BOLD:
                    return font
        
        # ignore weight
        return fonts[0] if loose else None
    
    
    def load_fonts(self, path):
        """
        Loads all supported fonts from specified folder.
        
        Args:
            path: str
                Absolute path to the fonts folder.
        """
        
        for root, dirs, files in os.walk(path):
            for file_name in files:
                self.load_font(os.path.join(root, file_name))
    
    
    def load_font(self, path, name=None):
        """
        Loads specified font file.
        
        Args:
            path: str
                Absolute path to the font definition file.
            
            name: str or None
                Specific (system) font name. Family name is used if not
                provided.
        """
        
        # get filename and extension
        dir_name, file_name = os.path.split(path)
        base_name, extension = os.path.splitext(file_name)
        
        # check font type
        if extension.lower() not in ('.ttf', '.otf', '.ttc'):
            return
        
        # init index
        index = 0
        
        # load all variants from file
        while True:
            
            # get font
            font = Font.from_ttf(path, name, index)
            if font is None:
                return
            
            # add to library
            if font.family not in self._fonts:
                self._fonts[font.family] = [font]
            else:
                self._fonts[font.family].append(font)
            
            # increase index
            index += 1
    
    
    def load(self):
        """Loads all supported fonts from known system locations."""
        
        # load MSWin fonts
        if sys.platform == 'win32':
            
            paths = [
                r"c:\Windows\Fonts"]
            
            for path in paths:
                if os.path.exists(path):
                    self.load_fonts(path)
        
        # load MacOS fonts
        elif sys.platform == 'darwin':
            
            paths = [
                r"/Library/Fonts/",
                r"/Network/Library/Fonts/",
                r"/System/Library/Fonts/"]
            
            home = os.environ.get('HOME')
            if home is not None:
                path = os.path.join(home, 'Library', 'Fonts')
                paths.append(path)
            
            for path in paths:
                if os.path.exists(path):
                    self.load_fonts(path)
        
        # load linux fonts
        elif sys.platform == 'linux':
            
            paths = [
                r"/usr/share/fonts",
                r"/usr/local/share/fonts",
                r"~/.local/share/fonts",
                r"/usr/share/fonts"]
            
            for path in paths:
                if os.path.exists(path):
                    self.load_fonts(path)
        
        # load iOS fonts
        elif sys.platform == 'ios':
            
            from objc_util import ObjCClass
            ui_font = ObjCClass('UIFont')
            
            for family in ui_font.familyNames():
                names = ui_font.fontNamesForFamilyName_(family)
                for name in names:
                    font = ImageFont.truetype(str(name), 10)
                    if font is not None:
                        self.load_font(font.path, str(name))


class Font(object):
    """This class holds some basic information about available font."""
    
    
    def __init__(self, path, index, font_name, font_family, font_type):
        """Initializes a new instance of Font."""
        
        self._cache = {}
        
        self._path = path
        self._index = index
        self._name = font_name
        self._family = font_family
        self._type = font_type
        
        self._style = FONT_STYLE_NORMAL
        self._weight = FONT_WEIGHT_NORMAL
        
        # set style
        if 'Italic' in font_type:
            self._style = FONT_STYLE_ITALIC
        
        elif 'Oblique' in font_type:
            self._style = FONT_STYLE_ITALIC
        
        # set weight
        if 'Regular' in font_type:
            self._weight = FONT_WEIGHT_NORMAL
        
        elif 'Bold' in font_type:
            self._weight = FONT_WEIGHT_BOLD
        
        elif 'Light' in font_type:
            self._weight = FONT_WEIGHT_LIGHT
        
        elif 'Black' in font_type:
            self._weight = FONT_WEIGHT_BLACK
        
        elif 'Heavy' in font_type:
            self._weight = FONT_WEIGHT_HEAVY
        
        elif 'Semibold' in font_type:
            self._weight = FONT_WEIGHT_SEMIBOLD
        
        elif 'Medium' in font_type:
            self._weight = FONT_WEIGHT_MEDIUM
        
        elif 'Ultralight' in font_type:
            self._weight = FONT_WEIGHT_ULTRALIGHT
        
        elif 'Thin' in font_type:
            self._weight = FONT_WEIGHT_THIN
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        return "%s (%s | %s | %s) [%s] - [%d]" % (self._family, self._name, self._style, self._weight, self._path, self._index)
    
    
    def __repr__(self):
        """Gets debug string representation."""
        
        return "%s(%s)" % (self.__class__.__name__, self.__str__())
    
    
    @property
    def path(self):
        """
        Gets absolute path to the font file.
        
        Returns:
            str
                Absolute path to font file.
        """
        
        return self._path
    
    
    @property
    def index(self):
        """
        Gets the font index within the file.
    
        Returns:
            int
                Font index.
        """
        
        return self._index
     
    
    @property
    def family(self):
        """
        Gets the font family name.
        
        Returns:
            str
                Font family name.
        """
        
        return self._family
    
    
    @property
    def name(self):
        """
        Gets the font name.
        
        Returns:
            str
                Font name.
        """
        
        return self._name
    
    
    @property
    def type(self):
        """
        Gets the font type.
        
        Returns:
            str
                Font type.
        """
        
        return self._type
    

    @property
    def style(self):
        """
        Gets the font style.
        
        Returns:
            pero.FONT_STYLE
                Font style.
        """
        
        return self._style
    
    
    @property
    def weight(self):
        """
        Gets the font weight.
        
        Returns:
            pero.FONT_WEIGHT
                Font weight.
        """
        
        return self._weight
    
    
    def get_font(self, size):
        """
        Gets initialized PIL font.
        
        Args:
            size: int
                Font size.
        
        Returns:
            PIL.ImageFont
                Initialized font with specified size.
        """
        
        if size not in self._cache:
            self._cache[size] = ImageFont.truetype(self._path, size, index=self._index)
        
        return self._cache[size]
    
    
    def get_size(self, text, size):
        """
        Gets text size for given text.
        
        Args:
            text: str
                Text for which the size should be calculated.
            
            size: int
                Font size to be used.
        
        Returns:
            (float, float)
                Text size as width and height.
        """
        
        font = self.get_font(size)
        
        return font.getbbox(text)[2:]
    
    
    def get_descent(self, size):
        """
        Gets font descent for given font size.
        
        Args:
            size: int
                Font size to be used.
        
        Returns:
            float
                Font descent.
        """
        
        font = self.get_font(size)
        
        m = font.getbbox("M")[3]
        mj = font.getbbox("Mj")[3]
        
        return mj - m
    
    
    @staticmethod
    def from_ttf(path, name=None, index=0):
        """
        Creates a new instance of Font.
        
        Args:
            path: str
                Absolute path to a font definition file.
            
            name: str
                Specific (system) name of the font variant.
            
            index: int
                Index of the font to load.
        
        Returns:
            pero.Font or None
                Font information or None if not supported.
        """
        
        try:
            font = ImageFont.truetype(path, 10, index=index)
            font_family, font_type = font.getname()
            font_name = name or font_family
            
            return Font(path, index, font_name, font_family, font_type)
        
        except:
            return None


# initializes available fonts
FONTS = FontManager()
FONTS.load()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . values import *
from . enum import Enum

# define backends
BACKEND_JSON = JSON
BACKEND_SVG = SVG
BACKEND_QT = QT
BACKEND_WX = WX
BACKEND_CAIRO = CAIRO
BACKEND_MUPDF = MUPDF
BACKEND_PY5 = PY5
BACKEND_PYTHONISTA = PYTHONISTA

BACKEND = Enum(
    JSON = BACKEND_JSON,
    SVG = BACKEND_SVG,
    QT = BACKEND_QT,
    WX = BACKEND_WX,
    CAIRO = BACKEND_CAIRO,
    MUPDF = BACKEND_MUPDF,
    PY5 = BACKEND_PY5,
    PYTHONISTA = BACKEND_PYTHONISTA)

# define image formats supported by JSON backend
EXPORT_JSON = {
    '.json'}

# define image formats supported by SVG backend
EXPORT_SVG = {
    '.svg'}

# define image formats supported by QT backend
EXPORT_QT = {
    '.bmp',
    '.gif',
    '.jpg',
    '.jpeg',
    '.pdf',
    '.png'}

# define image formats supported by WX backend
EXPORT_WX = {
    '.bmp',
    '.cur',
    '.ico',
    '.jpg',
    '.jpeg',
    '.pcx',
    '.png',
    '.pnm',
    '.tif',
    '.tiff',
    '.xpm'}

# define formats supported by Cairo backend
EXPORT_CAIRO = {
    '.bmp',
    '.eps',
    '.gif',
    '.jpg',
    '.jpeg',
    '.pdf',
    '.png',
    '.svg',
    '.tif',
    '.tiff'}

# define image formats supported by MuPDF backend
EXPORT_MUPDF = {
    '.pdf'}

# define image formats supported by PY5 backend
EXPORT_PY5 = {
    '.gif',
    '.jpeg',
    '.jpg',
    '.pdf',
    '.png',
    '.svg',
    '.tga'}

# define image formats supported by Pythonista backend
EXPORT_PYTHONISTA = {
    '.png'}

# define available formats
EXPORT_FORMATS = {
    BACKEND_JSON: EXPORT_JSON,
    BACKEND_SVG: EXPORT_SVG,
    BACKEND_QT: EXPORT_QT,
    BACKEND_WX: EXPORT_WX,
    BACKEND_CAIRO: EXPORT_CAIRO,
    BACKEND_MUPDF: EXPORT_MUPDF,
    BACKEND_PY5: EXPORT_PY5,
    BACKEND_PYTHONISTA: EXPORT_PYTHONISTA}

# define export backend priorities
EXPORT_PRIORITY = [
    BACKEND_JSON,
    BACKEND_SVG,
    BACKEND_QT,
    BACKEND_WX,
    BACKEND_CAIRO,
    BACKEND_MUPDF,
    EXPORT_PY5,
    BACKEND_PYTHONISTA]

# define viewer backend priorities
VIEWER_PRIORITY = [
    BACKEND_QT,
    BACKEND_WX,
    BACKEND_PY5,
    BACKEND_PYTHONISTA]

# define default export size
EXPORT_WIDTH = 750
EXPORT_HEIGHT = 500

VIEWER_WIDTH = 750
VIEWER_HEIGHT = 500



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. drawing import Graphics


class Glyph(Graphics):
    """Abstract base class for simple graphical entities."""
    
    pass



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from .. drawing import Matrix, Path
from . glyph import Glyph


class Head(Glyph):
    """
    Abstract base class for various types of heads, typically used as part of an
    arrow at the beginning and/or the end of the arrow line. They are simple
    glyphs defined by size and rotation angle, drawn at specified position.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the anchor.
        
        y: int, float or callable
            Specifies the y-coordinate of the anchor.
        
        size: int, float or callable
            Specifies the full size.
        
        angle properties:
            Includes pero.AngleProperties to specify the angle.
        
        line properties:
            Includes pero.LineProperties to specify the head outline.
        
        fill properties:
            Includes pero.FillProperties to specify the head fill.
    """
    
    TYPE = None
    
    x = NumProperty(0)
    y = NumProperty(0)
    size = NumProperty(10)
    angle = Include(AngleProperties)
    
    line = Include(LineProperties, line_color=UNDEF)
    fill = Include(FillProperties, fill_color=UNDEF)
    
    
    @staticmethod
    def create(symbol, **overrides):
        """
        Initializes a new instance of specified head.
        
        Args:
            symbol: str
                Head symbol.
        
        Returns:
            pero.Head
                Initialized head.
        """
        
        # convert from symbol
        if symbol == HEAD_CIRCLE:
            return CircleHead(**overrides)
        
        elif symbol == HEAD_LINE:
            return LineHead(**overrides)
        
        elif symbol in (HEAD_NORMAL, HEAD_NORMAL_B):
            return NormalHead(**overrides)
        
        elif symbol in (HEAD_OPEN, HEAD_OPEN_B):
            return OpenHead(**overrides)
        
        elif symbol in (HEAD_VEE, HEAD_VEE_B):
            return VeeHead(**overrides)
        
        raise ValueError("Unknown head symbol! -> '%s'" % symbol)


class CircleHead(Head):
    """Defines a circular head."""
    
    TYPE = HEAD_CIRCLE
    
    line_color = ColorProperty(None, nullable=True)
    fill_color = ColorProperty(UNDEF, nullable=True)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw head."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_circle(x, y, 0.5 * size)


class LineHead(Head):
    """Defines a strait line head."""
    
    TYPE = HEAD_LINE
    
    line_color = ColorProperty(UNDEF, nullable=True)
    fill_color = ColorProperty(None, nullable=True)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw head."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # get coords
        angle -= 0.5*math.pi
        x_shift = 0.5*size * math.cos(angle)
        y_shift = 0.5*size * math.sin(angle)
        
        # set pen
        canvas.set_pen_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_line(x-x_shift, y-y_shift, x+x_shift, y+y_shift)


class NormalHead(Head):
    """Defines a closed triangle head."""
    
    TYPE = HEAD_NORMAL
    
    line_color = ColorProperty(None, nullable=True)
    fill_color = ColorProperty(UNDEF, nullable=True)
    
    _f = 0.5/(math.sqrt(3)/2)
    _points = ((-_f*1.5, -0.5), (0,0), (-_f*1.5, +0.5))
    _path = Path().polygon(_points)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw head."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # transform path
        matrix = Matrix().scale(size, size).rotate(angle).translate(x, y)
        path = self._path.transformed(matrix)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_path(path)


class OpenHead(Head):
    """Defines an opened triangle head."""
    
    TYPE = HEAD_OPEN
    
    line_color = ColorProperty(UNDEF, nullable=True)
    fill_color = ColorProperty(None, nullable=True)
    
    _f = 0.5/(math.sqrt(3)/2)
    _points = ((-_f*1.5, -0.5), (0,0), (-_f*1.5, +0.5))
    _path = Path().move_to(*_points[0]).line_to(*_points[1]).line_to(*_points[2])
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw head."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # transform path
        matrix = Matrix().scale(size, size).rotate(angle).translate(x, y)
        path = self._path.transformed(matrix)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_path(path)


class SymbolHead(Head):
    """
    Defines a custom path head. The path must be a 'symbol-path' i.e. anchored
    at 0,0 coordinates, scaled to fit into 1x1 square and pointing right. This
    can be created from any path by calling the 'symbol' method.
    
    Properties:
        
        path: pero.Path, callable, None or UNDEF
            Specifies the symbol-path to be used as head.
    """
    
    path = Property(UNDEF, types=(Path,), nullable=True)
    
    line_color = ColorProperty(None, nullable=True)
    fill_color = ColorProperty(UNDEF, nullable=True)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw head."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        path = self.get_property('path', source, overrides)
        
        # check data
        if not path:
            return
        
        # transform symbol (assuming unit coords, zero center, pointing right)
        matrix = Matrix().scale(size, size).rotate(angle).translate(x, y)
        path = path.transformed(matrix)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_path(path)


class VeeHead(Head):
    """Defines a closed V-shape head."""
    
    TYPE = HEAD_VEE
    
    line_color = ColorProperty(None, nullable=True)
    fill_color = ColorProperty(UNDEF, nullable=True)
    
    _f = 0.5/(math.sqrt(3)/2)
    _points = ((-_f*1.5, -0.5), (0,0), (-_f*1.5, +0.5), (-1+_f, 0))
    _path = Path().polygon(_points)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw head."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # transform path
        matrix = Matrix().scale(size, size).rotate(angle).translate(x, y)
        path = self._path.transformed(matrix)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_path(path)


class HeadProperty(Property):
    """
    Defines a head property, which simplifies a head definition by converting
    specific symbols into an instance of corresponding head glyph. Available
    symbols are defined by the pero.HEAD enum.
    """
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of HeadProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (Head, str,)
        
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # check type
        if isinstance(value, Head):
            return value
        
        # parse main
        value = super().parse(value)
        
        # allow UNDEF and None
        if value is UNDEF or value is None:
            return value
        
        # check func
        if callable(value):
            return value
        
        # convert from symbol
        return Head.create(value)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# keep version info
QT_VERSION = None
QT_NAME = None

# import PyQt5
if QT_VERSION is None:
    
    try:
        from PyQt5.QtCore import Qt, QPoint, QLineF, QSizeF, QEvent
        from PyQt5.QtGui import QFont, QColor, QPen, QBrush, QPainterPath, QPageSize
        from PyQt5.QtGui import QPainter, QImage, QPicture, QPixmap, QFontMetrics
        from PyQt5.QtGui import QMouseEvent, QWheelEvent, QTouchEvent
        from PyQt5.QtWidgets import QApplication, QWidget, QHBoxLayout, QDesktopWidget
        from PyQt5.QtPrintSupport import QPrinter
        
        QT_VERSION = 5
        QT_NAME = 'PyQt5'
    
    except ImportError:
        pass

# import PySide2
if QT_VERSION is None:
    
    try:
        from PySide2.QtCore import Qt, QPoint, QLineF, QSizeF, QEvent
        from PySide2.QtGui import QFont, QColor, QPen, QBrush, QPainterPath, QPageSize
        from PySide2.QtGui import QPainter, QImage, QPicture, QPixmap, QFontMetrics
        from PySide2.QtGui import QMouseEvent, QWheelEvent, QTouchEvent
        from PySide2.QtWidgets import QApplication, QWidget, QHBoxLayout, QDesktopWidget
        from PySide2.QtPrintSupport import QPrinter
        
        QT_VERSION = 5
        QT_NAME = 'PySide2'
    
    except ImportError:
        pass

# import PyQt6
if QT_VERSION is None:
    
    try:
        from PyQt6.QtCore import Qt, QPoint, QLineF, QSizeF, QEvent
        from PyQt6.QtGui import QFont, QColor, QPen, QBrush, QPainterPath, QPageSize
        from PyQt6.QtGui import QPainter, QImage, QPicture, QPixmap, QFontMetrics
        from PyQt6.QtWidgets import QApplication, QWidget, QHBoxLayout
        from PyQt6.QtPrintSupport import QPrinter
        
        QT_VERSION = 6
        QT_NAME = 'PyQt6'
    
    except ImportError:
        pass

# import PySide6
if QT_VERSION is None:
    
    try:
        from PySide6.QtCore import Qt, QPoint, QLineF, QSizeF, QEvent
        from PySide6.QtGui import QFont, QColor, QPen, QBrush, QPainterPath, QPageSize
        from PySide6.QtGui import QPainter, QImage, QPicture, QPixmap, QFontMetrics
        from PySide6.QtWidgets import QApplication, QWidget, QHBoxLayout
        from PySide6.QtPrintSupport import QPrinter
        
        QT_VERSION = 6
        QT_NAME = 'PySide6'
    
    except IOError:
        pass

# check import
if QT_VERSION is None:
    raise ImportError()

# fix version 5
if QT_VERSION == 5:
    
    # set high DPI
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    
    # fix QWidget
    QWidget.screen = lambda s: QDesktopWidget()
    
    # fix QMouseEvt
    QMouseEvent.position = lambda s: s.pos()
    
    # fix QWheelEvent
    QWheelEvent.position = lambda s: s.pos()
    
    # fix QTouchEvent
    QTouchEvent.points = lambda s: s.touchPoints()
    QTouchEvent.TouchPoint.position = lambda d: d.pos()
    QTouchEvent.TouchPoint.lastPosition = lambda d: d.lastPos()
    
    # fix QPrinter
    QPrinter.setPageSize = lambda s, d: s.setPaperSize(d.definitionSize(), QPrinter.Unit.Point)


# fix QApplication 'exec_' method
if not hasattr(QApplication, 'exec'):
    QApplication.exec = lambda d: d.exec_()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from .. enums import *
from .. properties import *
from . scale import Scale


class OrdinalScale(Scale):
    """
    This type of scale tool is used to convert discrete values of input into
    predefined discrete values of output (e.g. from category names to colors).
    
    If an unknown input value is to be converted, the next available output
    value can be automatically assigned (and remembered since then) if allowed
    by setting the 'implicit' property to True, otherwise predefined 'default'
    value is returned. Note that if implicit assignment is allowed, newly
    assigned input values are kept within the 'in_range' property, which might
    cause some memory issues because they are never released.
    
    In cases there are more input values than predefined output values, the
    available output values can be recycled by setting the 'recycle' property to
    True. Otherwise predefined 'default' value is returned. This settings makes
    most sense if implicit assignment is allowed. Note that if recycling is
    allowed the 'invert' method in this case returns the first usage of provided
    value.
    
    Properties:
        
        default: any, None or UNDEF
            Specifies the default value.
        
        implicit: bool
            Specifies whether the next available output value should be
            automatically assigned to unknown input value.
        
        recycle: bool
            Specifies whether the output values can be reused if there are no
            free values remaining.
    """
    
    default = Property(UNDEF, dynamic=False, nullable=True)
    implicit = BoolProperty(False, dynamic=False)
    recycle = BoolProperty(False, dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of OrdinalScale."""
        
        super().__init__(**overrides)
        
        # init indexes
        self._in_range_idx = None
        self._out_range_idx = None
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_ordinal_scale_property_changed)
    
    
    def scale(self, value, *args, **kwargs):
        """
        Returns corresponding output value for given input value.
        
        Args:
            value: any or (any,)
                Input value to be scaled.
        
        Returns:
            any
                Scaled value.
        """
        
        # apply array scaling
        if isinstance(value, (numpy.ndarray, list, tuple)):
            return tuple(map(self.scale, value))
        
        # check output
        if len(self.out_range) == 0:
            return self.default
        
        # init index map
        if self._in_range_idx is None:
            self._in_range_idx = {}
            
            for i, item in enumerate(self.in_range):
                self._in_range_idx[item] = i
        
        # get index
        idx = self._in_range_idx.get(value, None)
        
        # allow implicit value for unknowns
        if idx is None and self.implicit:
            self.in_range = list(self.in_range) + [value]
            idx = len(self.in_range) - 1
        
        # use default value
        if idx is None:
            return self.default
        
        # check max index and recycle
        if idx >= len(self.out_range):
            if self.recycle:
                idx = idx % len(self.out_range)
            else:
                return self.default
        
        # return output value
        return self.out_range[idx]
    
    
    def invert(self, value, *args, **kwargs):
        """
        Returns corresponding input value for given output value.
        
        Args:
            value: any or (any,)
                Output value to be inverted.
        
        Returns:
            any
                Inverted value.
        """
        
        # apply array scaling
        if isinstance(value, (numpy.ndarray, list, tuple)):
            return tuple(map(self.invert, value))
        
        # init index map
        if self._out_range_idx is None:
            self._out_range_idx = {}
            
            for i, item in enumerate(self.out_range):
                self._out_range_idx[item] = i
        
        # get index
        idx = self._out_range_idx.get(value, None)
        
        # check index
        if idx is None or idx >= len(self.in_range):
            return None
        
        # return input value
        return self.in_range[idx]
    
    
    def _on_ordinal_scale_property_changed(self, evt=None):
        """Called after a property has changed."""
        
        # check in_range
        if evt is None or evt.name == 'in_range':
            self._in_range_idx = None
            if self.in_range is None or self.in_range is UNDEF:
                self.in_range = ()
        
        # check out_range
        if evt is None or evt.name == 'out_range':
            self._out_range_idx = None
            if self.out_range is None or self.out_range is UNDEF:
                self.out_range = ()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from .. enums import *
from .. properties import *


class Interpol(PropertySet):
    """
    Abstract base class for various types of interpolators. The interpolators
    are used within continuous scales to convert values between two dimensions.
    Such conversion is always done in two steps: normalizing given value within
    input range and de-normalizing the normalized value within output range. For
    such functionality each interpolator should implement specific methods for
    'normalize' and 'denormalize' values within provided ranges.
    
    In some cases like conversion of whole data series the input value is
    represented as a sequence of values rather than a single value. The
    interpolator therefore should be designed to accept such data type as well.
    """
    
    
    def normalize(self, x, start, end):
        """
        Calculates normalized value within specified range.
        
        Args:
            x: float or (float,)
                Value(s) to be normalized.
            
            start: float
                Start of the range to normalize into.
            
            end: float
                End of the range to normalize into.
        
        Returns:
            float or (float,)
                Normalized value(s).
        """
        
        raise NotImplementedError("The 'normalize' method is not implemented for '%s'." % self.__class__.__name__)
    
    
    def denormalize(self, x, start, end):
        """
        Calculates de-normalized value within specified range.
        
        Args:
            x: float or (float,)
                Normalized value(s) to be converted.
            
            start: float
                Start of the range to de-normalize from.
            
            end: float
                End of the range to de-normalize from.
        
        Returns:
            float or (float,)
                De-normalized value(s).
        """
        
        raise NotImplementedError("The 'normalize' method is not implemented for '%s'." % self.__class__.__name__)


class LinInterpol(Interpol):
    """
    This type of interpolator applies linear interpolation within specified
    ranges.
    """
    
    
    def normalize(self, x, start, end):
        """
        Calculates normalized value within specified range.
        
        Args:
            x: float or numpy.ndarray
                Value(s) to be normalized.
            
            start: float
                Start of the range to normalize into.
            
            end: float
                End of the range to normalize into.
        
        Returns:
            float or numpy.ndarray
                Normalized value(s).
        """
        
        return (x - start) / float(end - start)
    
    
    def denormalize(self, x, start, end):
        """
        Calculates de-normalized value within specified range.
        
        Args:
            x: float or numpy.ndarray
                Normalized value(s) to be converted.
            
            start: float
                Start of the range to de-normalize from.
            
            end: float
                End of the range to de-normalize from.
        
        Returns:
            float or numpy.ndarray
                De-normalized value(s).
        """
        
        return start + x * (end - start)
    

class LogInterpol(Interpol):
    """
    This type of interpolator applies logarithmic interpolation within specified
    ranges.
    """
    
    
    def normalize(self, x, start, end):
        """
        Calculates normalized value within specified range.
        
        Args:
            x: float or numpy.ndarray
                Value(s) to be normalized.
            
            start: float
                Start of the range to normalize into.
            
            end: float
                End of the range to normalize into.
        
        Returns:
            float, numpy.ndarray
                Normalized value(s).
        """
        
        return numpy.log(x / float(start)) / numpy.log(end / float(start))
    
    
    def denormalize(self, x, start, end):
        """
        Calculates de-normalized value within specified range.
        
        Args:
            x: float or numpy.ndarray
                Normalized value(s) to be converted.
            
            start: float
                Start of the range to de-normalize from.
            
            end: float
                End of the range to de-normalize from.
        
        Returns:
            value: float or numpy.ndarray
                De-normalized value(s).
        """
        
        return numpy.power(end, x) * numpy.power(start, 1-x)
    

class PowInterpol(Interpol):
    """
    This type of interpolator applies exponential interpolation within specified
    ranges.
    
    Properties:
        
        power: int or float
            Specifies the exponent to be used.
    """
    
    power = FloatProperty(1, dynamic=False)
    
    
    def normalize(self, x, start, end):
        """
        Calculates normalized value within specified range.
        
        Args:
            x: float or numpy.ndarray
                Value(s) to be normalized.
            
            start: float
                Start of the range to normalize into.
            
            end: float
                End of the range to normalize into.
        
        Returns:
            float or numpy.ndarray
                Normalized value(s).
        """
        
        power = self.power
        a = numpy.power(start, power)
        b = numpy.power(end, power) - a
        
        return (numpy.power(x, power) - a) / float(b)
    
    
    def denormalize(self, x, start, end):
        """
        Calculates de-normalized value within specified range.
        
        Args:
            x: float or numpy.ndarray
                Normalized value(s) to be converted.
            
            start: float
                Start of the range to de-normalize from.
            
            end: float
                End of the range to de-normalize from.
        
        Returns:
            float or numpy.ndarray
                De-normalized value(s).
        """
        
        power = self.power
        a = numpy.power(start, power)
        b = numpy.power(end, power) - a
        c = 1./power
        
        return numpy.power(a + x * b, c)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for time formatter."""
    
    
    def test_standalone(self):
        """Tests whether formatter works as standalone tool."""
        
        # init formatter
        formatter = pero.TimeFormatter()
        
        # test hours
        self.assertEqual(formatter.format(3600), "1.00")
        self.assertEqual(formatter.format(3600+1800), "1.50")
        
        # test minutes
        self.assertEqual(formatter.format(60), "1.00")
        self.assertEqual(formatter.format(3600-1), "59.98")
        
        # test seconds
        self.assertEqual(formatter.format(1), "1.00")
        self.assertEqual(formatter.format(17.555), "17.55")
        self.assertEqual(formatter.format(59), "59.00")
        
        # test milliseconds
        self.assertEqual(formatter.format(.001), "1.00")
        self.assertEqual(formatter.format(.1), "100.00")
    
    
    def test_units(self):
        """Tests whether units are added correctly."""
        
        formatter = pero.TimeFormatter(add_units=True)
        
        self.assertEqual(formatter.format(3600+1800-1), "1.50 h")
        self.assertEqual(formatter.format(70.5), "1.18 m")
        self.assertEqual(formatter.format(17.567), "17.57 s")
        self.assertEqual(formatter.format(.1), "100.00 ms")
    
    
    def test_precision(self):
        """Tests whether precision is used correctly."""
        
        # check hours
        formatter = pero.TimeFormatter(precision=3600)
        
        self.assertEqual(formatter.format(3600+1800-1), "1.50")
        self.assertEqual(formatter.format(70.5), "1.18")
        self.assertEqual(formatter.format(17.567), "17.57")
        self.assertEqual(formatter.format(.1), "100.00")
        
        # check minutes
        formatter = pero.TimeFormatter(precision=60)
        
        self.assertEqual(formatter.format(3600+1800-1), "1:30")
        self.assertEqual(formatter.format(70.5), "1.18")
        self.assertEqual(formatter.format(17.567), "17.57")
        self.assertEqual(formatter.format(.1), "100.00")
        
        # check seconds
        formatter = pero.TimeFormatter(precision=1)
        
        self.assertEqual(formatter.format(3600+1800-1), "1:29:59")
        self.assertEqual(formatter.format(70.5), "01:10")
        self.assertEqual(formatter.format(17.567), "17.57")
        self.assertEqual(formatter.format(.1), "100.00")
        
        # check milliseconds
        formatter = pero.TimeFormatter(precision=0.1)
        
        self.assertEqual(formatter.format(3600+1800-1), "1:29:59:000")
        self.assertEqual(formatter.format(70.5), "01:10:500")
        self.assertEqual(formatter.format(17.567), "17:567")
        self.assertEqual(formatter.format(.1), "100.00")
    
    
    def test_domain(self):
        """Tests whether domain is used correctly."""
        
        # test hours
        formatter = pero.TimeFormatter(domain=3600)
        
        self.assertEqual(formatter.format(3600+1800-1), "1.50")
        self.assertEqual(formatter.format(70.5), "0.02")
        self.assertEqual(formatter.format(17.567), "0.00")
        self.assertEqual(formatter.format(.1), "0.00")
        
        # test minutes
        formatter = pero.TimeFormatter(domain=60)
        
        self.assertEqual(formatter.format(3600+1800-1), "89.98")
        self.assertEqual(formatter.format(70.5), "1.18")
        self.assertEqual(formatter.format(17.567), "0.29")
        self.assertEqual(formatter.format(.1), "0.00")
        
        # test seconds
        formatter = pero.TimeFormatter(domain=1)
        
        self.assertEqual(formatter.format(3600+1800-1), "5399.00")
        self.assertEqual(formatter.format(70.5), "70.50")
        self.assertEqual(formatter.format(17.567), "17.57")
        self.assertEqual(formatter.format(.1), "0.10")
    
    
    def test_auto_range(self):
        """Tests whether domain together with precision are used correctly."""
        
        formatter = pero.TimeFormatter(domain=3600, precision=1)
        
        self.assertEqual(formatter.format(3600+1800-1), "1:29:59")
        self.assertEqual(formatter.format(70.5), "0:01:10")
        self.assertEqual(formatter.format(17.567), "0:00:18")
        self.assertEqual(formatter.format(.1), "0:00:00")
    
    
    def test_separator(self):
        """Tests whether separator is used correctly."""
        
        formatter = pero.TimeFormatter(separator=" ", add_units=True, domain=3600, precision=1)
        
        self.assertEqual(formatter.format(3600+1800-1), "1 h 29 m 59 s")
        self.assertEqual(formatter.format(70.5), "0 h 01 m 10 s")
        self.assertEqual(formatter.format(17.567), "0 h 00 m 18 s")
        self.assertEqual(formatter.format(.1), "0 h 00 m 00 s")
    
    
    def test_template(self):
        """Tests whether custom template is used correctly."""
        
        formatter = pero.TimeFormatter(
            template = "time: {h}:{m}:{s}.{ms}",
            domain = 3600,
            precision = 1,
            add_units = True)
        
        self.assertEqual(formatter.format(3600+1800-1), "time: 1:29:59.000")
        self.assertEqual(formatter.format(70.5), "time: 0:01:10.500")
        self.assertEqual(formatter.format(17.567), "time: 0:00:17.567")
        self.assertEqual(formatter.format(.1), "time: 0:00:00.100")
    
    
    def test_parts_templates(self):
        """Tests whether custom parts templates are used correctly."""
        
        # with specified unit
        formatter = pero.TimeFormatter(
            template="{h} {m} {s} {ms} {us} {ns}",
            h_template = "h:{h:.0f}",
            m_template = "m:{m:.0f}",
            s_template = "s:{s:.0f}",
            ms_template = "ms:{ms:.0f}",
            us_template = "us:{us:.0f}",
            ns_template = "ns:{ns:.0f}")
        
        self.assertEqual(formatter.format(3600+1800-1), "h:1 m:29 s:59 ms:0 us:0 ns:0")
        self.assertEqual(formatter.format(70.5), "h:0 m:1 s:10 ms:500 us:0 ns:0")
        self.assertEqual(formatter.format(17.567), "h:0 m:0 s:17 ms:567 us:0 ns:0")
        self.assertEqual(formatter.format(.1), "h:0 m:0 s:0 ms:100 us:0 ns:0")
        
        # without specified unit
        formatter = pero.TimeFormatter(
            template="{h} {m} {s} {ms} {us} {ns}",
            h_template = "h:{:.0f}",
            m_template = "m:{:.0f}",
            s_template = "s:{:.0f}",
            ms_template = "ms:{:.0f}",
            us_template = "us:{:.0f}",
            ns_template = "ns:{:.0f}")
        
        self.assertEqual(formatter.format(3600+1800-1), "h:1 m:29 s:59 ms:0 us:0 ns:0")
        self.assertEqual(formatter.format(70.5), "h:0 m:1 s:10 ms:500 us:0 ns:0")
        self.assertEqual(formatter.format(17.567), "h:0 m:0 s:17 ms:567 us:0 ns:0")
        self.assertEqual(formatter.format(.1), "h:0 m:0 s:0 ms:100 us:0 ns:0")
    
    
    def test_rounding(self):
        """Tests whether rounding works correctly."""
        
        # round down
        formatter = pero.TimeFormatter(rounding=pero.ROUNDING.FLOOR, add_units=True)
        
        self.assertEqual(formatter.format(3600+1800-1), "1.49 h")
        self.assertEqual(formatter.format(70.5), "1.17 m")
        self.assertEqual(formatter.format(17.567), "17.56 s")
        
        # round up
        formatter = pero.TimeFormatter(rounding=pero.ROUNDING.CEIL, add_units=True)
        
        self.assertEqual(formatter.format(3600+1800-1), "1.50 h")
        self.assertEqual(formatter.format(70.5), "1.18 m")
        self.assertEqual(formatter.format(17.567), "17.57 s")


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)
